from typing import Optional, Union, Sequence, NamedTuple, Tuple, Any, Dict, List, Type, Iterator, Callable
import sys
from pathlib import Path
from tabulate import tabulate
from collections import OrderedDict
import inspect
from omnibelt import unspecified_argument, Class_Registry, Function_Registry, colorize, include_module

from .. import __logger__ as prt
from ..abstract import AbstractConfig, AbstractProject, AbstractBehavior, AbstractCustomArtifact
from ..config import ConfigManager



class ProjectBase(AbstractProject):
	'''
	Generally, each workspace (e.g. repo, directory, package) should define its own project,
	which keeps track of all the configs and artifacts defined therein.

	Projects don't just contain all the project specific registries, but can also modify the behavior
	of the top-level methods such as (such as :func:`main()`, :func:`run()`, etc.).

	It is recommended to subclass this class to create a custom project class with expected functionality
	(unlike :class:`AbstractProject`) and automatically register it to the global project type registry.
	'''

	global_type_registry = Class_Registry()
	'''Global registry for project types (usually used by the profile to create custom project types).'''

	def __init_subclass__(cls, name: Optional[str] = None, **kwargs):
		super().__init_subclass__()
		cls._type_name = name
		if name is not None:
			cls.global_type_registry.new(name, cls)


	@classmethod
	def get_project_type(cls, ident: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''Accesses the project type entry for the given identifier (from a registry).'''
		return cls.global_type_registry.find(ident, default=default)



class GeneralProject(ProjectBase, name='general'):
	'''Project class that includes basic functionality such as a script registry and config manager.'''

	info_file_names = {'.fig.project.yml', '.fig.project.yaml', 'fig.project.yml', 'fig.project.yaml',
	                   '.omnifig.yml', '.omnifig.yaml', 'omnifig.yml', 'omnifig.yaml'}


	def infer_path(self, path: Optional[Union[str, Path]] = None) -> Path:
		'''Infers the path of the project from the given path or the current working directory.'''
		if isinstance(path, str):
			path = Path(path)
		if path is None:
			path = Path()
		if path.is_file():
			return path
		if path.is_dir():
			for name in self.info_file_names:
				p = path / name
				if p.exists():
					return p
		# prt.warning(f'Could not infer project path from {path} (using blank project)')


	def _format_xray_entry(self, entry: NamedTuple) -> Tuple[List[str], Optional[str]]:
		'''Prints a single artifact entry (e.g. script) with pretty formatting (including color)'''

		desc = getattr(entry, 'description', None)

		key_fmt = colorize('{key}', color='green')

		proj = getattr(entry, 'project', None)

		item = getattr(entry, 'cls', None)
		if item is None:
			item = getattr(entry, 'fn', None)
		if isinstance(item, AbstractCustomArtifact):
			item = item.get_wrapped()
		bases = getattr(item, '__bases__', None)

		path = getattr(entry, 'path', None)

		terms = [
			proj.name if proj is not None else '',

		    key_fmt.format(key=entry.name)
		]

		if bases is not None:
			terms.append(colorize(item.__name__, color="blue"))

			terms.append(item.__module__)

			terms.append(f'<{", ".join([b.__name__ for b in item.__bases__])}>')

		elif item is not None:
			terms.append(colorize(item.__name__, color="blue"))

			terms.append(item.__module__)

			terms.append(f'{inspect.signature(item)}')

		elif path is not None: # config

			root = getattr(proj, 'root', None)

			if root is not None:
				try:
					extra = str(path.relative_to(root))
				except:
					extra = str(path)
				terms.append(extra)

			else:
				terms.append('')

		else:
			raise NotImplementedError(f'Unknown entry type: {entry}')

		return terms, desc


	def xray(self, artifact: str, *, sort: Optional[bool] = False, reverse: Optional[bool] = False,
	         as_list: Optional[bool] = False) -> Optional[List[NamedTuple]]:
		'''
		Prints a list of all artifacts of the given type accessible from this project
		(including related and active base projects).

		Args:
			artifact: artifact type (e.g. 'script', 'config')
			sort: sort the list of artifacts by name
			reverse: reverse the order of the list of artifacts
			as_list: instead of printing, return the list of artifacts

		Returns:
			list: if as_list is True, returns a list of artifacts

		Raises:
			UnknownArtifactTypeError: if the given artifact type does not exist

		'''
		terms = list(self.iterate_artifacts(artifact))

		if sort:
			terms.sort(key=lambda x: x.name)
		if reverse:
			terms.reverse()

		if as_list:
			return terms

		if len(terms):
			rows, descs = zip(*[self._format_xray_entry(t) for t in terms])
			descs = [None if d is None else f'\t[{d}]' for d in descs]

			table = tabulate(rows, tablefmt='simple')

			lines = [line for lines in zip(table.splitlines()[1:-1], descs) for line in lines if line is not None]
			print('\n'.join(lines))


	class Script_Registry(Function_Registry, components=['description', 'hidden', 'project']):
		'''Registry for scripts (functions) that can be run from the command line.'''
		pass


	Config_Manager = ConfigManager


	def __init__(self, path: Optional[Union[str, Path]], *, script_registry: Optional[Script_Registry] = None,
	             config_manager: Optional[Config_Manager] = None, **kwargs):
		if script_registry is None:
			script_registry = self.Script_Registry()
		if config_manager is None:
			config_manager = self.Config_Manager(self)
		path = self.infer_path(path)
		super().__init__(path, **kwargs)
		self._behaviors = None
		self._path = None if path is None else path.absolute()
		self._modules = OrderedDict()
		self.config_manager = config_manager
		self._artifact_registries = {
			'script': script_registry,
		}


	def load(self):
		'''
		Loads all dependencies for the project including config files/directories, packages, and source files
		based on the contents of the project's info file.

		Packages are imported (or reloaded) and source files are executed locally.

		Returns:
			Project: self

		'''
		if self.root is not None:
			# load any dependencies
			dep_dir = self.data.get('auto_dependency_dir', 'dependency')
			if dep_dir is not None:
				deps = self.data.get(dep_dir, [])
				if isinstance(deps, str):
					deps = [deps]
				for dep in deps:
					self._profile.get_project(dep).load()

			# config files/directories
			self.load_configs(self.data.get('configs', []))
			config_dir = self.data.get('auto_config_dir', 'config')
			if config_dir is not None:
				path = self.root / config_dir
				if path.is_dir():
					self.load_configs([path])

			# import modules
			module_dir = self.data.get('auto_module_dir', 'module')
			modules = []
			if module_dir is not None:
				modules = self.data.get(module_dir, [])
				if isinstance(modules, str):
					modules = [modules]

			# run source files
			src_dir = self.data.get('auto_src_dir', 'src')
			src = []
			if src_dir is not None:
				src = self.data.get(src_dir, [])
				if isinstance(src, str):
					src = [src]

			self._run_dependencies(src, modules)


	def unload(self):
		'''Unloads all dependencies for the project (opposite of :meth:`load`).'''
		for module, (mod, deps) in self._modules.items():
			for name, dep in deps.items():
				if name in sys.modules:
					del sys.modules[name]


	def _activate(self, *args, **kwargs):
		'''
		Activates the project including the info from the data file
		which includes registering all configs (keys: ``config`` and ``configs``)
		and source files (keys: ``packages``, ``package``, ``src``).
		'''
		super()._activate(*args, **kwargs)
		prt.info(f'Activating project {self.name} ({self.root})')
		with self._profile.project_context(self):
			self.load()


	def _deactivate(self, *args, **kwargs):
		'''Deactivates the project including unloading all dependencies.'''
		super()._deactivate(*args, **kwargs)
		prt.info(f'Deactivating project {self.name} ({self.root})')
		self.unload()

	def load_configs(self, paths: Sequence[Union[str ,Path]] = ()) -> None:
		'''Registers all specified config files and directories'''
		for path in paths:
			path = Path(path)
			if path.is_file():
				raise NotImplementedError
			elif path.is_dir():
				self.register_config_dir(path, recursive=True)


	def _run_dependencies(self, srcs: Optional[Sequence[Union[str , Path]]] = (),
	                      packages: Optional[Sequence[str]] = ()) -> None:
		'''Imports all specified packages and runs the specified python files'''
		modules = [*map(Path,srcs), *packages]
		for module in modules:
			# if module in self._modules:
			# 	mod, dependencies = self._modules[module]
			# 	sys.modules.update(dependencies)
			# else:
			mod, dependencies = include_module(module, root=self.root)

			root = str(Path(mod.__file__).parent.resolve())

			sub = {name: dep for name, dep in dependencies.items()
				   if getattr(dep, '__file__', None) is not None and str(Path(dep.__file__).resolve()).startswith(root)}
			self._modules[module] = mod, sub

		for owner, submodules in self._modules.values():
			for name, sub in submodules.items():
				del sys.modules[name] # remove any local modules from sys.modules

	# region Organization
	def extract_info(self, other: 'GeneralProject') -> None:
		'''Extracts the info from the given project into this project.'''
		super().extract_info(other)
		self._path = other._path
		self._profile = other._profile
		self.config_manager = other.config_manager
		self._artifact_registries = other._artifact_registries


	def validate(self) -> AbstractProject:
		'''Validates the project and returns a new project with the validated info.'''
		requested_type = self.data.get('type', None)
		if requested_type is not None and requested_type != self._type_name:
			prt.info(f'Replacing project type {self._type_name!r} with {requested_type!r}')
			entry = self._profile.get_project_type(requested_type)
			proj = entry.cls(self.info_path, profile=self._profile)
			proj.extract_info(self)
			return proj
		return self


	def __repr__(self):
		name = self.name
		return f'{self.__class__.__name__}[{"default" if name is None else name}]({self.root})'


	def __eq__(self, other):
		if not isinstance(other, GeneralProject):
			return False
		return self.info_path == other.info_path


	def __hash__(self):
		return hash(self.info_path)


	@property
	def name(self):
		'''The name of the project'''
		return self.data.get('name', None)
	@name.setter
	def name(self, value: str):
		self.data['name'] = value


	@property
	def root(self) -> Path:
		'''The root directory of the project.'''
		if self._path is not None:
			return self._path.parent


	@property
	def info_path(self) -> Path:
		'''The path to the info file.'''
		return self._path
	# endregion


	# region Running/Ops
	def create_config(self, *parents: str, **parameters):
		'''Creates a config with the given parameters using the config manager.'''
		return self.config_manager.create_config(parents, parameters)


	def parse_argv(self, argv: Sequence[str], *, script_name: Optional[str] = unspecified_argument) -> AbstractConfig:
		'''Parses the given command line arguments into a config object.'''
		return self.config_manager.parse_argv(argv, script_name=script_name)


	def behaviors(self) -> Iterator[AbstractBehavior]:
		'''Iterates over all behaviors associated with this project.'''
		# return self._profile.iterate_behaviors()
		if self._behaviors is None:
			self._behaviors = [entry.cls(self) for entry in self._profile.iterate_behaviors()]
		yield from sorted(self._behaviors, reverse=True)


	def validate_run(self, config: AbstractConfig) -> Optional[AbstractProject]:
		'''
		Validates the project ``self`` using the given config object before running it.

		More specifically, this method calls :meth:`validate_project` on all behaviors associated with the project
		with the config object as argument. If any of the behaviors returns a new project, this project
		is returned (and used for the script execution instead). Otherwise, ``None`` is returned.

		Args:
			config: The config object to use for validation.

		Returns:
			The new project to use for the script execution or ``None`` if no new project was returned.

		'''
		for behavior in self.behaviors():
			out = behavior.validate_project(config)
			if out is not None:
				return out


	def main(self, argv: Sequence[str], script_name: Optional[str] = unspecified_argument) -> Any:
		'''
		Runs the script with the given arguments using the config object obtained by parsing ``argv``.

		More specifically, this method does the following:
			1. Activates the project (loads any specified source files or packages for the script, if not done yet).
			2. Instantiates all behaviors associated with the project.
			3. Parses the command line arguments into a config object.
			4. Validates the current project using the config object and behaviors (see :meth:`validate_run`).
			5. Runs the script using the config object (see :meth:`run_script`).
			6. Cleans up the project (see :meth:`cleanup`).

		Args:
			argv: List of top-level arguments (expected to be :code:`sys.argv[1:]`).
			script_name: specified name of the script
			(defaults to what is specified in argv when it is parsed into a config object).

		Returns:
			The output of the script.

		'''
		self.activate()  # load the project

		self._behaviors = None  # clear the behaviors cache
		behaviors = [behavior for behavior in self.behaviors()]  # instantiate all behaviors

		config = self.parse_argv(argv, script_name=script_name)

		transfer = self.validate_run(config)  # can update/modify the project based on the config
		if transfer is not None:
			return transfer.main(argv, script_name=script_name)

		output = self.run_script(script_name, config)  # run based on the config
		self.cleanup()  # cleanup
		return output  # return output


	def run_script(self, script_name: str, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Runs the script with the given arguments using :func:`run()` of the current project.

		Args:
			script_name: The script name to run (must be registered).
			config: Config object to run the script with (must include the script under :code:`_meta.script_name`).
			args: Additional positional arguments to pass to the script.
			kwargs: Additional keyword arguments to pass to the script.

		Returns:
			The output of the script.

		'''
		if script_name not in {None, unspecified_argument}:
			config.push('_meta.script_name', script_name, overwrite=True, silent=True) # TODO: handle readonly configs
		return self.run(config, *args, **kwargs)


	class NoScriptError(ValueError):
		'''Raised when the script name is not specified.'''
		pass

	def run(self, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Runs the given script with the given config using this project.

		Args:
			config: The config to use for running the script (must include the script under :code:`_meta.script_name`).
			args: Additional positional arguments to pass to the script.
			kwargs: Additional keyword arguments to pass to the script.

		Returns:
			The return value of the script.

		'''
		config.project = self
		meta = config.push_peek('_meta', {}, overwrite=False, silent=True) # TODO: handle readonly configs

		behaviors = [behavior for behavior in self.behaviors() if behavior.include(meta)]

		# pre run
		for behavior in behaviors:
			try:
				out = behavior.pre_run(meta, config)
			except self.TerminationFlag as e:
				return e.out
			except:
				prt.error(f'Error while running behavior {behavior!r} pre run')
				raise
			else:
				if out is not None:
					config = out

		script_name = meta.pull('script_name', None, silent=True)
		if script_name is None:
			raise self.NoScriptError
		entry = self.find_script(script_name)
		try:
			output = self._run(entry, config, args=args, kwargs=kwargs) # run script
		except Exception as exc:
			try:
				for behavior in behaviors:
					behavior.handle_exception(meta, config, exc)
			except self.TerminationFlag as e:
				return e.out
			except self.IgnoreException as e:
				output = e.out
			else:
				raise exc

		# post run
		for behavior in behaviors:
			try:
				out = behavior.post_run(meta, config, output)
			except self.TerminationFlag as e:
				return e.out
			except:
				prt.error(f'Error while running behavior {behavior!r} post run')
				raise
			else:
				if out is not None:
					output = out

		return output


	TerminationFlag = AbstractBehavior.TerminationFlag
	IgnoreException = AbstractBehavior.IgnoreException


	def _run(self, script_entry: Script_Registry.entry_cls, config: AbstractConfig, *,
	         args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None) -> Any:
		'''
		Runs the given script with the given config.

		Args:
			script_entry: The script entry to run.
			config: The config to use for running the script.
			args: Additional positional arguments to pass to the script.
			kwargs: Additional keyword arguments to pass to the script.

		Returns:
			The return value of the script.

		'''
		if args is None:
			args = ()
		if kwargs is None:
			kwargs = {}

		fn = script_entry.fn
		if isinstance(fn, AbstractCustomArtifact):
			fn = fn.top
		return fn(config, *args, **kwargs)
	# endregion


	# region Registration
	class UnknownArtifactTypeError(KeyError):
		'''Raised when an unknown artifact type is encountered.'''
		pass


	def find_local_artifact(self, artifact_type: str, ident: str,
	                        default: Optional[Any] = unspecified_argument) -> Optional[NamedTuple]:
		'''
		Finds the artifact of the given type and registered with the given identifier.

		Args:
			artifact_type: The type of artifact to find.
			ident: The identifier of the artifact to find.
			default: The default value to return if the artifact is not found.

		Returns:
			The artifact entry from the registry corresponding to the given type.

		Raises:
			UnknownArtifactTypeError: If the artifact type is not registered.
			UnknownArtifactError: If the artifact is not found and no default is given.

		'''
		self.activate()
		if artifact_type == 'config':
			try:
				return self.config_manager.find_local_config_entry(ident, default=default)
			except self.config_manager.ConfigNotRegistered:
				pass
		else:
			registry = self._artifact_registries.get(artifact_type)
			if registry is None:
				raise self.UnknownArtifactTypeError(artifact_type)
			try:
				return registry.find(ident, default=unspecified_argument)
			except registry.NotFoundError:
				pass
		if default is not unspecified_argument:
			return default
		raise self.UnknownArtifactError(artifact_type, ident)


	def find_artifact(self, artifact_type: str, ident: str,
	                  default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds an artifact in the project's registries, including related and active base projects.
		Artifacts are data or functionality such as configs and components.

		Args:
			artifact_type: Type of artifact to find (eg. 'config' or 'component').
			ident: Name of the artifact that was registered.
			default: Default value to return if the artifact is not found.

		Returns:
			Artifact object, or, if a default value is given and artifact is not found.

		Raises:
			:class:`UnknownArtifactError`: if the artifact is not found and no default is specified.

		'''
		return self.find_local_artifact(artifact_type, ident, default=default)


	def register_artifact(self, artifact_type, ident: str, artifact: Union[str, Type, Callable], *,
	                      project: Optional[AbstractProject] = unspecified_argument, **kwargs) -> NamedTuple:
		'''
		Registers the given artifact with the given type and identifier and any additional info.

		Args:
			artifact_type: The type of artifact to register.
			ident: The identifier of the artifact to register.
			artifact: The artifact to register (usually a type or config file path).
			project: The project to register the artifact for (defaults to this project).
			kwargs: Additional keyword arguments to pass to the registry.

		Returns:
			The artifact entry from the registry corresponding to the given type.

		Raises:
			UnknownArtifactTypeError: If the artifact type does not exist.

		'''
		if project is unspecified_argument or project is None:
			project = self
		if artifact_type == 'config':
			return self.config_manager.register_config(ident, artifact, project=project)
		registry = self._artifact_registries.get(artifact_type)
		if registry is None:
			raise self.UnknownArtifactTypeError(artifact_type)
		return registry.new(ident, artifact, project=project, **kwargs)


	def iterate_artifacts(self, artifact_type: str) -> Iterator[NamedTuple]:
		'''
		Iterates over all artifacts of the given type.

		Args:
			artifact_type: The type of artifact to iterate over (e.g. 'script', 'config').

		Returns:
			An iterator over all artifacts of the given type.

		Raises:
			UnknownArtifactTypeError: If the artifact type does not exist.

		'''
		self.activate()
		if artifact_type == 'config':
			yield from self.config_manager.iterate_configs()
		elif artifact_type not in self._artifact_registries:
			raise self.UnknownArtifactTypeError(artifact_type)
		else:
			yield from self._artifact_registries[artifact_type].values()


	def find_config(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the config with the given name.

		Args:
			name: name the config was registered with.
			default: default value to return if the config is not found.

		Returns:
			The config entry corresponding to the given name.

		Raises:
			UnknownArtifactError: If the config is not found and no default is given.

		'''
		return self.find_artifact('config', name, default=default)


	def register_config(self, name: Union[str, Path], path: Union[str, Path] = None, **kwargs) -> NamedTuple:
		'''
		Registers a config file with the given name.

		Note:
			It is generally not recommended to register configs manually, but rather to use the ``register_config_dir``
			method to register all configs in a directory at once.

		Args:
			name: to register the config under
			path: of the config file (if not provided, the provided name is assumed to be a path)
			**kwargs: Other arguments to pass to the ``Path_Registry.register`` method

		Returns:
			The entry of the config file that was registered

		'''
		return self.register_artifact('config', name, path, **kwargs)


	def iterate_configs(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered config file entries.

		Returns:
			An iterator over all registered config file entries.

		'''
		yield from self.iterate_artifacts('config')


	def register_config_dir(self, path: Union[str, Path], *, recursive: Optional[bool] = True,
	                        prefix: Optional[str] = None, delimiter: Optional[str] = None) -> List[NamedTuple]:
		'''
		Registers all yaml files found in the given directory (possibly recursively)

		When recusively checking all directories inside, the internal folder hierarchy is preserved
		in the name of the config registered, so for example if the given ``path`` points to a
		directory that contains a directory ``a`` and two files ``f1.yaml`` and ``f2.yaml``:

		Contents of ``path`` and corresponding registered names:

			- ``f1.yaml`` => ``f1``
			- ``f2.yaml`` => ``f2``
			- ``a/f3.yaml`` => ``a/f3``
			- ``a/b/f4.yaml`` => ``a/b/f3``

		If a ``prefix`` is provided, it is appended to the beginning of the registered names

		Args:
			path: path to root directory to search through
			recursive: search recursively through subdirectories for more config yaml files
			prefix: prefix for names of configs found herein
			delimiter: string to merge directories when recursively searching (default ``/``)

		Returns:
			A list of all config entries that were registered.

		'''
		return self.config_manager.register_config_dir(path, recursive=recursive, prefix=prefix, delimiter=delimiter)


	def find_script(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the script with the given name.

		Args:
			name: the script was registered with.
			default: default value to return if the script is not found.

		Returns:
			The script entry corresponding to the given name.

		Raises:
			UnknownArtifactError: If the script is not found and no default is given.

		'''
		return self.find_artifact('script', name, default=default)


	def register_script(self, name: str, fn: Callable[[AbstractConfig], Any], *, description: Optional[str] = None,
	                    hidden: Optional[bool] = None) -> NamedTuple:
		'''
		Register a script with the given name.

		Args:
			name: to register the script under
			fn: the script function (should expect the first positional argument to be the config object)
			description: description of the script
			hidden: whether to hide the script from the list of available scripts
			(defaults to whether the name starts with an underscore)

		Returns:
			The entry of the script that was registered

		'''
		return self.register_artifact('script', name, fn, description=description, hidden=hidden)


	def iterate_scripts(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered script entries.

		Returns:
			An iterator over all registered script entries.

		'''
		yield from self.iterate_artifacts('script')
	# endregion
	pass





from typing import Any, Dict, List, Optional

from .. import __logger__ as prt
from ..abstract import AbstractBehavior, AbstractConfig
from ..top import get_profile



class Behavior(AbstractBehavior):
	'''Recommended parent class for meta rules.'''

	name = None
	code = None
	priority = 0
	num_args = 0
	description = None

	def __init_subclass__(cls, name: Optional[str] = None, description: Optional[str] = None,
	                      code: Optional[str] = None, priority: Optional[int] = None,
	                      num_args: Optional[int] = None, **kwargs):
		super().__init_subclass__(**kwargs)

		# default to class attributes
		if name is None:
			name = getattr(cls, 'name', None)
		if code is None:
			code = getattr(cls, 'code', None)
		if priority is None:
			priority = getattr(cls, 'priority', 0)
		if num_args is None:
			num_args = getattr(cls, 'num_args', 0)
		if description is None:
			description = getattr(cls, 'description', None)

		# infer missing
		if code is not None and name is None:
			prt.warning(f'No name for {Behavior.__name__} {cls.__name__} provided, will default to {cls.__name__!r}')
			name = cls.__name__
		if code is None and name is not None:
			prt.error(f'No code for {Behavior.__name__} {name!r} provided, '
			          f'cannot register a {Behavior.__name__} without a code')

		# registration
		if name is not None:
			cls.name = name
			cls.code = code
			cls.num_args = num_args
			cls.priority = priority
			cls.description = description
			get_profile().register_behavior(name, cls, description=description)


	def __gt__(self, other):
		'''Compares the behavior to another behavior for ordering.'''
		return self.priority > other.priority


	def __lt__(self, other):
		'''Compares the behavior to another behavior for ordering.'''
		return self.priority < other.priority


	def parse_argv(self, meta: Dict[str, Any], argv: List[str],
	               script_name: Optional[str] = None) -> Optional[List[str]]:
		'''
		Optionally modifies the arguments when the project's :meth:`main()` is called.

		Args:
			meta: Meta-data extracted from the argv so far (can be modified here).
			argv: List of arguments to parse (expected to be :code:`sys.argv[1:]`).
			script_name: Manually specified name of the script (if not provided, it will be parsed from argv).

		Returns:
			Modified list of arguments (or :code:`None` if no modification is needed).

		'''

		if self.code is not None:
			term = f'-{self.code}'
			if term in argv: # trigger this behavior
				idx = argv.index(term)
				assert len(argv) > idx + self.num_args, f'Expected {self.num_args} argument/s after {term!r} for' \
				                                        f'behavior {self.name!r} but only {len(argv) - idx - 1} ' \
				                                        f'were provided'

				meta[self.name] = argv[idx + 1:idx + 1 + self.num_args] if self.num_args > 0 else True
				if self.num_args > 0:
					argv = argv[:idx] + argv[idx + 1 + self.num_args:]
				argv = [a for a in argv if a != term] # remove all instances of the term
				return argv


	def include(self, meta: AbstractConfig) -> bool:
		'''
		Checks if the current behavior should be included before running the given config.

		Args:
			config: Config object to use.

		Returns:
			:code:`True` if the behavior should be included, :code:`False` otherwise.

		'''
		return self.name in meta and (self.num_args > 0 or meta.pull(self.name, False, silent=True))
















from typing import Union, Any, Type, Optional
from pathlib import Path
import yaml
from omnibelt import SimpleExporterBase, ExportManager, get_now
from omnibelt.exporting import ExporterBase, AbstractExportManager
from omnibelt import exporting_common as _

# from .abstract import AbstractConfig
# from .config.nodes import ConfigSparseNode, ConfigDenseNode
from .config import ConfigNode
from .top import create_config



class ConfigExporter(SimpleExporterBase, extensions=['.fig.yml', '.fig.yaml'], types=ConfigNode):
	'''
	Exporter for config objects, can load and save config objects in four formats: json, yaml, toml,
	and the native ".fig.yml" format (which is equivalent to yaml).
	'''

	def export_payload(self, src: AbstractExportManager, payload: ConfigNode, path: Path, *,
					   fmt: Optional[str] = None, **kwargs) -> Path:
		'''
		Exports the given payload to the given path.

		Args:
			src: manager used for delegating the export to a different format
			payload: config object to be exported
			path: destination path for the export
			fmt: format to use for the export (if None, will be inferred from the path, defaults to `.fig.yml`)

		Returns:
			the path to the exported file (or None if the export failed)

		'''
		if path.suffix.endswith('.toml') or path.suffix.endswith('.tml'):
			return src.export(payload.to_python(), path=path, fmt='toml')
		elif path.suffix.endswith('.json'):
			return src.export(payload.to_python(), path=path, fmt='json')
		else:
			return super().export_payload(src, payload, path, fmt=fmt, **kwargs)


	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		lines = [
			f'# Exported on {get_now()}',
		]

		cro = getattr(payload, 'cro', None)
		if cro is not None:
			lines.append(f'# Config resolution order (cro): {cro}')

		bases = getattr(payload, 'bases', None)
		if bases is not None:
			lines.append(f'# Config bases: {bases}')

		lines.append(payload.to_yaml())
		with path.open('w') as f:
			f.write('\n'.join(lines))


	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		'''
		Load a config object from a file. Will use the config manager of the current project,
		see ``ConfigManager.create_config``.

		Args:
			path: the path to the file to load
			**kwargs: additional arguments to pass to the load

		Returns:
			the loaded config object

		'''
		return create_config(path, **kwargs)












from typing import Any, Dict, List, Optional, Tuple, Union, Type, Sequence, Callable
import inspect
from omnibelt import dynamic_capture, extract_function_signature, Modifiable

from .abstract import AbstractConfig, AbstractConfigurable, AbstractCertifiable



class Configurable(AbstractConfigurable, Modifiable):
	'''
	Mix-in class for objects that can be constructed with a config object.

	It is strongly recommended that components and modifiers inherit from this class
	to seamlessly fill in missing arguments with the config object.
	'''
	_my_config = None
	# @property
	# def my_config(self) -> AbstractConfig:
	# 	'''The config object that this object is associated with'''
	# 	return getattr(self, '_my_config', None)


	# @classmethod
	# def inject_mods(cls, *mods, name=None):
	# 	'''
	# 	Mods should be types that are used to modify the cls. The order of the mods corresponds to the order
	# 	in which the cls is "modified", so for example:
	#
	# 	```python
	#
	# 	class A(Configurable): pass
	# 	class B: pass
	# 	class C: pass
	#
	# 	out = A.inject_mods(C, B)
	# 	assert out.mro() == [out, C, B, A, Modifiable, object]
	# 	assert out.__name__ == 'C_B_A' # default name
	# 	```
	# 	'''
	# 	return super().inject_mods(*reversed(mods), name=name)


	class _config_builder_type:
		'''Replaces the regular method and fills in the missing arguments from the config'''
		def __init__(self, product: 'Configurable', config: AbstractConfig, *, silent: Optional[bool] = None):
			self.product = product
			self.config = config
			self.silent = silent


		@staticmethod
		def configurable_parents(cls: Type) -> List[Type['Configurable']]:
			'''Returns a list of all the parent classes that are Configurable'''
			return [c for c in cls.mro() if issubclass(c, Configurable) and c is not Configurable]


		class MissingConfigError(Exception):
			'''Raised when a method is missing arguments from the config'''
			def __init__(self, product: Any, missing: List[inspect.Parameter], config: Optional[AbstractConfig] = None,
			             msg: Optional[str] = None):
				config_key = ''
				if config is not None:
					trace = config.trace
					if trace is not None:
						config_key = f'<{config.reporter.get_key(trace)}> '
				missing = [repr(p.name) for p in missing]
				if msg is None:
					s = 's' if len(missing) > 1 else ''
					msg = f'{config_key}{product.__name__} is missing {len(missing)} arg{s}: {", ".join(missing)}'
				super().__init__(msg)
				self.product = product
				self.missing = missing
				self.config = config


		def find_missing_arg(self, name: str, default: Optional[inspect.Parameter] = inspect.Parameter.empty) -> Any:
			'''
			Finds the missing argument in the config (including aliases)

			Args:
				name: Name of the argument
				default: Default specified in the method signature

			Returns:
				The value of the argument from the config

			'''
			param = None
			if default is not inspect.Parameter.empty:
				assert isinstance(default, inspect.Parameter)
				param = default
				default = param.default
			if default is inspect.Parameter.empty:
				default = self.config._empty_default
			aliases = self.aliases.get(name, ())
			silent = self.silent
			if (self.silences is not None and name in self.silences) \
				or (param is not None and (name == 'args' and param.kind == inspect.Parameter.VAR_POSITIONAL) 
										or (name == 'kwargs' and param.kind == inspect.Parameter.VAR_KEYWORD)):
				silent = True
				if default is self.config._empty_default:
					default = () if name == 'args' else {}
			out = self.config.pulls(name.replace('_', '-'), *aliases, default=default, silent=silent)
			return out


		def fix_args(self, method: Callable, obj: Any, args: Tuple, kwargs: Dict[str, Any]) \
				-> Tuple[Tuple, Dict[str, Any]]:
			'''
			Fills in the missing arguments from the config given the method signature

			Args:
				method: Method to fill in the arguments for
				obj: Object that the method is being called on
				args: Manually specified arguments
				kwargs: Manually specified keyword arguments

			Returns:
				The arguments and keyword arguments filled in from the config and the manually specified ones

			'''
			fixed_args, fixed_kwargs, missing = extract_function_signature(method, (obj, *args), kwargs,
			                                                               default_fn=self.find_missing_arg,
			                                                               include_missing=True)
			if len(missing):
				raise self.MissingConfigError(type(obj), missing, config=self.config)
			return fixed_args, fixed_kwargs


		def fixer(self, owner: Type, method: Callable, obj: Any, args: Tuple, kwargs: Dict[str, Any]) -> Any:
			'''
			Called by the capture to replace the original method call

			See ``omnibelt.tricks.dynamic_capture`` for more details.

			Args:
				owner: Class where the method is defined
				method: Method to call
				obj: Owner object (self in the method)
				args: Manually specified arguments
				kwargs: Manually specified keyword arguments

			Returns:
				The return value of the method

			'''
			self.aliases = getattr(method, '_my_config_aliases', {})
			self.silences = getattr(method, '_my_silent_config', None)

			fixed_args, fixed_kwargs = self.fix_args(method, obj, args, kwargs)
			return method(*fixed_args, **fixed_kwargs)


		def build(self, *args, **kwargs):

			init_capture = dynamic_capture(self.configurable_parents(self.product), self.fixer, '__init__').activate()

			self.product._my_config = self.config

			obj = self.product(*args, **kwargs)

			del self.product._my_config
			obj._my_config = self.config

			init_capture.deactivate()
			return obj


	@classmethod
	def _config_builder(cls, config, silent=None):
		return cls._config_builder_type(cls, config, silent=silent)


	@classmethod
	def init_from_config(cls, config: AbstractConfig,
	                     args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None, *,
	                     silent: Optional[bool] = None) -> Any:
		'''
		Constructor to initialize a class informed by the config object `config`. This will run the usual constructor
		``__init__``, except any arguments that are missing from the signature will be filled in from the config.

		Args:
			config: Config object to use
			args: Manually specified arguments
			kwargs: Manually specified keyword arguments
			silent: If True, no messages are reported when querying the config object.

		Returns:
			The initialized object

		'''
		if args is None:
			args = ()
		if kwargs is None:
			kwargs = {}
		return cls._config_builder(config, silent=silent).build(*args, **kwargs)



class Certifiable(Configurable, AbstractCertifiable):
	'''
	Simple mix-in to make the initialization of classes through the config a two-stage process.
	The first stage calls the ``__init__`` method, and then after that is complete, the ``__certify__`` method
	is called, which can return a new object to replace the original one.

	Note, that ``__certify__`` is only called if the object is initialized through the config (e.g. through ``pull()``).
	'''
	def __certify__(self, config: AbstractConfig, **kwargs):
		return self



class silent_config_args:
	'''Decorator to silence the config when extracting arguments from the config'''
	def __init__(self, *args: str):
		self.args = args


	def __call__(self, fn: Callable) -> Callable:
		if not hasattr(fn, '_my_silent_config'):
			setattr(fn, '_my_silent_config', set())
		fn._my_silent_config.update(self.args)
		return fn



class config_aliases:
	'''Method decorator to add aliases to the config arguments of a method'''
	def __init__(self, **aliases: Union[Sequence[str], str]):
		'''
		Config aliases to store for the method

		Args:
			**aliases: Mapping of aliases: keys should be the name of the argument in the method signature,
			and the values should be the name of the argument to query in the config object.
		'''
		self.aliases = aliases


	def __call__(self, fn: Callable) -> Callable:
		'''Decorator to add aliases to the config arguments of a method'''
		if not hasattr(fn, '_my_config_aliases'):
			setattr(fn, '_my_config_aliases', {})
		fn._my_config_aliases.update({key: ((val,) if isinstance(val, str) else tuple(val))
		                              for key, val in self.aliases.items()})
		return fn








import omnifig as fig
from . import util


class Module:
	def forward(self, *args, **kwargs):
		raise NotImplementedError

	def __call__(self, *args, **kwargs):
		return self.forward(*args, **kwargs)

	def parameters(self):
		return []


@fig.component('nn')
class NeuralNetwork(fig.Configurable, Module):
	@fig.config_aliases(input_dim='input_shape', output_dim='output_shape')
	def __init__(self, input_dim, output_dim, layers, nonlin='relu', use_gpu=True):
		on_gpu = ' (on gpu)' if use_gpu else ''
		self.info = f'Net({input_dim} -> {nonlin} layers:{layers}{on_gpu} -> {output_dim})'
		self.layers = layers

	def forward(self, x):
		return len(self.layers) * x

	def parameters(self):
		return [self.info]


@fig.component('net')
class DeepModel(fig.Configurable, Module):
	@fig.config_aliases(input_dim='input_shape', output_dim='output_shape')
	def __init__(self, input_dim, output_dim, # arguments without a default must appear in the config object
	             layers=[64, 64], name=None, use_gpu=False, nonlin='relu',
	             logger=None, **kwargs):  # some arguments may include subcomponents that are automatically created
		super().__init__(**kwargs)
		self.name = name
		self.use_gpu = use_gpu

		# although it is not recommended, you can of course also manually create components
		# (but this will hard code their use)
		self.net = NeuralNetwork(input_dim, output_dim, layers, nonlin=nonlin, use_gpu=use_gpu)

		self.logger = logger
	
	def forward(self, x):
		return self.net(x)
	
	def parameters(self):
		return self.net.parameters()


@fig.component('sgd-optim')
class Optim(fig.Configurable):
	def __init__(self, *args, learning_rate=1e-3, **kwargs):
		self.learning_rate = learning_rate
		self.parameters = []
	
	def include_parameters(self, parameters):
		self.parameters.extend(parameters)

	def step(self, y, pred):
		return 0.001




from typing import Optional, Iterator, NamedTuple, List
import sys
from tabulate import tabulate
from omnibelt import colorize

from ..abstract import AbstractConfig, AbstractCustomArtifact
from .. import get_current_project, get_profile

from .base import Behavior



class Help(Behavior, name='help', code='h', priority=99, num_args=0, description='Display this help message'):
	'''
	When activated, this behavior prints the help message for the current project (and then exits the program).
	'''

	@staticmethod
	def _format_script_entry(entry: NamedTuple, verbose: bool = False) -> List[str]:
		'''Formats a script entry for printing help message.'''
		ident = entry.name
		desc = getattr(entry, 'description', '')

		proj = getattr(entry, 'project', None)

		item = getattr(entry, 'cls', None)
		if item is None:
			item = getattr(entry, 'fn', None)
		if isinstance(item, AbstractCustomArtifact):
			item = item.get_wrapped()

		name = colorize(getattr(item, '__name__', ''), 'blue')
		mod = getattr(item, '__module__', None)
		if len(name) and mod is not None:
			name = f'{mod}.{name}'

		owner = getattr(proj, 'name', '')

		row = [
			colorize(ident, color='green'),
			name if verbose else mod,

			desc,
		]
		if verbose:
			row.insert(2, owner)
		return row


	_script_title = '''
Usage: {usage}
	
This will run the script: {script} {script_info}
	
{doc}

You can additionally specify the following:'''


	@classmethod
	def format_selected_script(cls, config: AbstractConfig, entry: Optional[NamedTuple], verbose: bool = False) -> str:
		'''Formats the help message about the selected script.'''
		if entry is not None:

			terms = cls._format_script_entry(entry, verbose=True)
			ident, name, owner, desc = terms

			item = getattr(entry, 'cls', None)
			if item is None:
				item = getattr(entry, 'fn', None)
			if isinstance(item, AbstractCustomArtifact):
				item = item.get_wrapped()

			info = {'script': ident}

			if item.__doc__ is not None:
				doc = item.__doc__
				doc = [line if line.startswith('\t') else f'\t{line}' for line in doc.splitlines()]
				doc = '\n'.join(doc)
			elif desc is not None:
				doc = '\t' + desc
			else:
				doc = '-- no description or documentation found --'
			info['doc'] = doc

			if name is not None:
				info['script_info'] = f'(found in {name})'

			info['usage'] = cls._format_usage(entry.name)

			return {'header': cls._script_title.format(**info)}


	@classmethod
	def format_scripts(cls, config: AbstractConfig, entries: Iterator[NamedTuple], verbose: bool = False) -> str:
		'''Formats the help message about the available scripts.'''
		rows = [cls._format_script_entry(s, verbose=verbose) for s in entries]

		if not len(rows):
			return '  -- No scripts registered --'

		cols = ['code', 'name' if verbose else 'module', 'description']
		if verbose:
			cols.insert(2, 'project')

		return tabulate(rows, headers=cols, tablefmt='simple')


	@staticmethod
	def _format_behavior_entry(entry: NamedTuple, verbose: bool = False) -> List[str]:
		'''Formats a behavior entry for printing help message.'''
		ident = entry.name
		desc = getattr(entry, 'description', '')

		item = getattr(entry, 'cls', None)
		if item is None:
			item = getattr(entry, 'fn', None)
		if isinstance(item, AbstractCustomArtifact):
			item = item.get_wrapped()

		name = colorize(getattr(item, '__name__', ''), 'blue')
		mod = getattr(item, '__module__', None)
		if len(name) and mod is not None:
			name = f'{mod}.{name}'

		code = colorize(getattr(item, 'code', ''), 'green')
		if len(code):
			code = f'-{code}'

		row = [
			code,
			ident,

			desc,
		]
		if verbose:
			row.insert(2, name)
		return row


	@classmethod
	def format_behaviors(cls, config: AbstractConfig, entries: Iterator[NamedTuple], verbose: bool = False) -> str:
		'''Formats the help message about the available behaviors.'''
		rows = [cls._format_behavior_entry(entry, verbose=verbose) for entry in entries]

		if len(rows) == 0:
			return '  -- No configs registered --'

		cols = ['code', 'name', 'description']
		if verbose:
			cols.insert(2, 'class')
		table = tabulate(rows, headers=cols, tablefmt='simple')
		return table


	@staticmethod
	def format_configs(config: AbstractConfig, entries: Iterator[NamedTuple], verbose: bool = False) -> str:
		'''Formats the help message about the available configs.'''
		names = [f'{e.name}' for e in entries]
		terms = [f'Config{"" if len(names) == 1 else "s"} ({len(names)}){":" if len(names) else ""}',
		         ', '.join(sorted(names))]
		if len(names) == 0:
			terms.clear()
			terms.append('  -- No configs registered --')

		return ' '.join(terms)


	_blank_title = '''
Usage: {usage}

+--------+
| script |
+--------+
Specify a registered script name to run

{scripts}'''


	_msg_template = '''{header}

+------+
| meta |
+------+
Optional meta behaviors to modify the execution 
(must be specified separately e.g. "-a -b")

{behaviors}

+---------+
| configs |
+---------+
Specify any registered config files or paths to YAML files 
to compose and pass to the script

{configs}

+------+
| args |
+------+
Any additional arguments specified manually
as key-value pairs (either as "--key value" or "--key=value")

Current project: {project}'''

	_default_usage = 'fig [-<meta>] <script> [<configs>...] [--<args>]'


	@classmethod
	def _format_usage(cls, name: Optional[str] = None):
		raw = sys.argv

		terms = ['fig' if len(raw) == 0 or raw[0].endswith('fig') else f'python {raw[0]}']

		if name in raw:
			terms.append(name)

		terms.append(
			'[-<meta>] <script> [<configs>...] [--<args>]' if name is None else '[-<meta>] [<configs>...] [--<args>]'
		)

		return colorize(' '.join(terms), 'magenta')


	@classmethod
	def pre_run(cls, meta: AbstractConfig, config: AbstractConfig) -> None:
		'''
		When activated, this behavior prints the help message for the current project (and then exits the program).

		Args:
			meta: The meta config object (used to check if the behavior is activated with the ``quiet`` key)
			config: The config object to be modified

		Returns:
			``None``

		Raises:
			:exc:`TerminationFlag`: if the rule is activated, to prevent the script from running

		'''

		with config.silence():
			verbose = config.pull('verbose', False)
			name = meta.pull('script_name', None)

		profile = get_profile()
		project = get_current_project()


		msg_info = {
			'project': project.name,
			'behaviors': cls.format_behaviors(config, profile.iterate_behaviors(), verbose=verbose),
			'configs': cls.format_configs(config, project.xray('config', as_list=True), verbose=verbose),
		}

		if name is not None:
			entry = project.find_script(name, None)
			if entry is not None:
				script_info = cls.format_selected_script(config, entry, verbose=verbose)
				if script_info is not None:
					msg_info.update(script_info)

		if 'header' not in msg_info:

			scripts = cls.format_scripts(config, project.xray('script', as_list=True), verbose=verbose)

			msg_info['header'] = cls._blank_title.format(scripts=scripts, usage=cls._format_usage(name))


		print(cls._msg_template.format(**msg_info))

		raise cls.TerminationFlag



import omnifig as fig
from _test_util import reset_profile



def test_configurable1():
	reset_profile()
	fig.initialize('example9')


	cfg = fig.create_config('c1', a=20)

	obj = cfg.pull('o1')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Cmpn1'
	assert obj.a == 20
	assert obj.b == 2

	cfg = fig.create_config('c1')

	obj = cfg.pull('o1')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Cmpn1'
	assert obj.a == 10
	assert obj.b == 2



def test_configurable2():
	reset_profile()
	fig.initialize('example9')

	C = fig.create_config('c2')

	obj = C.pull('obj1')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Cmpn3'
	assert obj.a == -1
	assert obj.b == -2
	assert obj.c == 5

	obj = C.pull('obj2')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Cmpn3'
	assert obj.a == -1
	assert obj.b == 100
	assert obj.c == 4

	obj = C.peek('obj3').process(a=-100)
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Cmpn3'
	assert obj.a == -100
	assert obj.b == -2
	assert obj.c == -3

	obj = C.pull('obj4')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Cmpn2'
	assert obj.a == 0
	assert obj.b == 20
	assert obj.c == 30



def test_mod_configurable():
	reset_profile()
	fig.initialize('example9')

	C = fig.create_config('c3')

	obj = C.peek_process('obj1', None, -2)
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Mod1_Cmpn1'
	assert obj.a == -2
	assert obj.b == 99
	assert obj.c == 5

	obj = C.pull('obj2')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Mod3_Cmpn1'
	assert obj.a == 9
	assert obj.b == 10
	assert obj.c == 9

	obj = C.pull('obj3')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Mod3_Cmpn1'
	assert obj.a == -88
	assert obj.b == 10
	assert obj.c == 5

	obj = C.pull('obj4')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Mod2_Mod3_Cmpn1'
	assert obj.a == -22
	assert obj.b == 10
	assert obj.c == 5

	obj = C.pull('obj5')
	assert isinstance(obj, fig.Configurable)
	assert type(obj).__name__ == 'Mod3_Mod2_Cmpn1'
	assert obj.a == -88
	assert obj.b == 8
	assert obj.c == 5



def test_key_error():
	@fig.component('something')
	class Something(fig.Configurable):
		def __init__(self, a=1, b={}):
			self.a = a
			self.b = b
			print(b['nonexistent'])

	cfg = fig.create_config(_type='something')

	try:
		cfg.create()
	except KeyError as e:
		assert e.args[0] == 'nonexistent'
	else:
		assert False, 'KeyError not raised'






from typing import List, Dict, Tuple, Optional, Union, Any, Sequence, Callable, Type, \
	Iterator, NamedTuple, ContextManager
# import abc
from pathlib import Path
from omnibelt import unspecified_argument, Registry, Primitive, JSONABLE

from .mixins import Activatable, FileInfo



class AbstractConfig:
	'''Abstract class for config objects'''

	# TODO: copy, deepcopy

	_empty_default = unspecified_argument


	class SearchFailed(KeyError):
		'''Raised when a search fails'''
		pass


	@property
	def cro(self) -> Tuple[str, ...]:
		'''
		Returns the list of all config files that were composed to produce this config tree.
		Analogous to the method resolution order (``mro``) for classes.
		'''
		raise NotImplementedError


	@property
	def bases(self) -> Tuple[str, ...]:
		'''
		Returns the list of config files that were explicitly mentioned to produce this config tree.
		Analogous to ``__bases__`` for classes.
		'''
		raise NotImplementedError


	@property
	def project(self) -> 'AbstractProject':
		'''Returns the project object associated with this config'''
		raise NotImplementedError
	@project.setter
	def project(self, project: 'AbstractProject'):
		'''Sets the project object associated with this config'''
		raise NotImplementedError


	@property
	def root(self) -> 'AbstractConfig':
		'''Returns the root node of the config object'''
		raise NotImplementedError


	def export(self, name: Union[str, Path], *, root: Optional[Union[str, Path]] = None,
			   fmt: Optional[str] = None) -> Optional[Path]:
		'''
		Exports the config object to the given path.

		Args:
			name: Name/path to export the config object to.
			root: Path to use as the root of the config object (defaults to the current working directory).
			fmt: Format to export the config object in (if None, will be inferred from the path).

		'''
		raise NotImplementedError


	def peek(self, query: Optional[str] = None, default: Optional[Any] = _empty_default, *,
	         silent: Optional[bool] = None) -> 'AbstractConfig':
		'''
		Returns a config object given by searching for the query in the config object.

		Args:
			query: The query to search for. If None, returns self.
			default: The default value to return if the query is not found.
			silent: If True, no message is reported based on the search.

		Returns:
			The config object found by the query, or the default value if the query is not found,
			and returns self if no query is provided.

		'''
		return self.peeks(query, default=default, silent=silent)


	def pull(self, query: Optional[str] = None, default: Optional[Any] = _empty_default, *,
	         silent: Optional[bool] = None) -> Any:
		'''
		Returns the value found by searching for the query in the config object.

		Args:
			query: The query to search for. If None, returns default.
			default: The default value to return if the query is not found.
			silent: If True, no message is reported based on the search.

		Returns:
			The value found by the query, or the default value if the query is not found,
			and returns the value of self if no query is provided.

		'''
		return self.pulls(query, default=default, silent=silent)


	def push_pull(self, addr: str, value: Any, *, overwrite: bool = True, silent: Optional[bool] = None) -> Any:
		'''
		Composes the push and pull methods into a single method. Can be used to set a value if none exists,
		and otherwise pull the existing value.

		Args:
			addr: The address to push/pull from.
			value: The value to push to the address.
			overwrite: If True, the value is pushed to the address even if it already exists.
			silent: If True, no message is reported.

		Returns:
			The value found at the address.

		'''
		self.push(addr, value, overwrite=overwrite, silent=silent)
		return self.pull(addr, silent=silent)


	def push_peek(self, addr: str, value: Any, *, overwrite: bool = True, silent: Optional[bool] = None) \
			-> 'AbstractConfig':
		'''
		Composes the push and peek methods into a single method. Can be used to set a value if none exists,
		and otherwise peek the existing value.

		Args:
			addr: The address to push/peek from.
			value: The value to push to the address.
			overwrite: If True, the value is pushed to the address even if it already exists.
			silent: If True, no message is reported.

		Returns:
			The value found at the address.

		'''
		self.push(addr, value, overwrite=overwrite, silent=silent)
		return self.peek(addr, silent=silent)


	def peeks(self, *queries, default: Optional[Any] = _empty_default,
	          silent: Optional[bool] = None) -> 'AbstractConfig':
		'''
		Returns a config object given by searching for the queries in the config object.
		If multiple queries are provided, each query is searched if the previous query fails.

		Args:
			*queries: Any number of queries to search for.
			default: The default value to return if none of the queries are not found.
			silent: If True, no message is reported based on the search.

		Returns:
			The config object found by the query, or the default value if none of the queries are found.

		'''
		raise NotImplementedError


	def pulls(self, *queries: str, default: Optional[Any] = _empty_default, silent: Optional[bool] = None) -> Any:
		'''
		Returns the value found by searching for the queries in the config object.

		Args:
			*queries: Any number of queries to search for.
			default: The default value to return if none of the queries are not found.
			silent: If True, no message is reported based on the search.

		Returns:
			The value found by the query, or the default value if none of the queries are found.

		'''
		raise NotImplementedError


	def push(self, addr: str, value: Any, overwrite: bool = True, *, silent: Optional[bool] = None) -> bool:
		'''
		Sets the value at the given address.

		Args:
			addr: The address to set the value at.
			value: The value to set.
			overwrite: If True, the value is set even if it already exists.
			silent: If True, no message is reported.

		Returns:
			True if the value was set, False if the value was not set.

		'''
		raise NotImplementedError


	def update(self, update: 'AbstractConfig') -> 'AbstractConfig':
		'''
		Updates the config object with the given config object (recursively overwriting common keys).

		Args:
			update: Config object to update with.

		Returns:
			The updated config object (self).

		'''
		raise NotImplementedError


	def silence(self, silent: bool = True) -> ContextManager:
		'''
		Context manager to silence the config object.

		Args:
			silent: If True, no messages are reported when querying the config object.

		Returns:
			Context manager to silence the config object.

		'''
		raise NotImplementedError


	def peek_named_children(self, *, silent: Optional[bool] = None) -> Iterator[Tuple[str, 'AbstractConfig']]:
		'''
		Returns an iterator of the child config objects of self together with their keys.

		Args:
			silent: If True, no messages are reported.

		Returns:
			An iterator of the child config objects of self together with their keys.

		'''
		raise NotImplementedError


	def peek_children(self, *, silent: Optional[bool] = None) -> Iterator['AbstractConfig']:
		'''
		Returns an iterator of the child config objects of self.

		Args:
			silent: If True, no messages are reported.

		Returns:
			An iterator of the child config objects of self.

		'''
		raise NotImplementedError


	def pull_named_children(self, *, force_create: Optional[bool] = False, silent: Optional[bool] = None) \
			-> Iterator[Tuple[str, Any]]:
		'''
		Returns an iterator of the child values of self together with their keys.

		Args:
			force_create: If True, creates new child values even if they already exist.
			silent: If True, no messages are reported.

		Returns:
			An iterator of the child values of self together with their keys.

		'''
		raise NotImplementedError


	def pull_children(self, *, force_create: Optional[bool] = False, silent: Optional[bool] = None) -> Iterator[Any]:
		'''
		Returns an iterator of the child values of self.

		Args:
			force_create: If True, creates new child values even if they already exist.
			silent: If True, no messages are reported.

		Returns:
			An iterator of the child values of self.

		'''
		raise NotImplementedError


	def create(self, *args: Any, **kwargs: Any) -> Any:
		'''
		Creates a new value based on the contents of self.

		Args:
			*args: Manual arguments to pass to the value constructor.
			**kwargs: Manual keyword arguments to pass to the value constructor.

		Returns:
			The newly created value.

		'''
		raise NotImplementedError


	def create_silent(self, *args: Any, **kwargs: Any) -> Any:
		'''
		Creates a new value based on the contents of self silently.

		Args:
			*args: Manual arguments to pass to the value constructor.
			**kwargs: Manual keyword arguments to pass to the value constructor.

		Returns:
			The newly created value.

		'''
		raise NotImplementedError


	def peek_create(self, query, default: Optional[Any] = _empty_default, *args, **kwargs):
		'''
		Composes the peek and create methods into a single method.

		Can be used to create a value using the config object found with the query.

		Args:
			query: The query to search for.
			default: If provided, the value is created using the default config object.
			*args: Manual arguments to pass to the value constructor.
			**kwargs: Manual keyword arguments to pass to the value constructor.

		Returns:
			The newly created value or `default` if the query is not found.

		'''
		raise NotImplementedError


	def process(self, *args: Any, **kwargs: Any) -> Any:
		'''
		Processes the config object using the contents of self.

		If a value for this config object has already been created, it is returned instead of creating a new one.

		Args:
			*args: Manual arguments to pass to the value constructor. (ignored if a value has already been created)
			**kwargs: Manual keyword arguments to pass to the value constructor.
			(ignored if a value has already been created)

		Returns:
			The processed value.

		'''
		raise NotImplementedError


	def peek_process(self, query, default: Optional[Any] = _empty_default, *args, **kwargs):
		'''
		Composes the peek and process methods into a single method.

		Args:
			query: The query to search for.
			default: If provided, the value is processed using the default config object.
			*args: Manual arguments to pass to the value constructor. (ignored if a value has already been created)
			**kwargs: Manual keyword arguments to pass to the value constructor.
			(ignored if a value has already been created)

		Returns:
			The processed value or `default` if the query is not found.

		'''
		raise NotImplementedError



class AbstractConfigurable:
	'''Abstract mix-in for objects that can be constructed using a config object.'''

	@classmethod
	def init_from_config(cls, config: AbstractConfig, args: Tuple = None, kwargs: Dict[str, Any] = None, *,
	                     silent: Optional[bool] = None) -> 'AbstractConfigurable':
		'''
		Constructor to initialize a class informed by the config object `config`.

		It is recommended that this should be a parent of any class that is registered as a component or modifier.

		Args:
			config: Config object to fill in any needed parameters.
			args: Manually specified arguments to pass to the constructor.
			kwargs: Manually specified keyword arguments to pass to the constructor.
			silent: If True, no messages are reported when querying the config object.

		Returns:
			Initialized instance of this class.

		'''
		raise NotImplementedError



class AbstractCertifiable(AbstractConfigurable):
	'''Abstract mix-in for objects that can must be certified after intialization.'''

	def __certify__(self, config: AbstractConfig) -> Optional[Any]:
		'''
		Certifies the object.

		Args:
			silent: If True, no messages are reported.

		'''
		return self



class AbstractConfigManager:
	'''Abstract class for config managers.'''


	class ConfigNotRegistered(KeyError):
		'''A config was not registered'''
		pass


	ConfigNode = None


	def register_config(self, name: str, path: Union[str, Path]) -> NamedTuple:
		'''
		Registers a path as a config file with the given name.

		Args:
			name: Name to register the config file under.
			path: Path to the config file.

		Returns:
			The entry that was registered (should contain at least the attributes for the name and path).

		'''
		raise NotImplementedError


	def register_config_dir(self, root: Union[str, Path]) -> List[NamedTuple]:
		'''
		Recursively registers all config files found in the given directory.

		Args:
			root: Path to the directory of the config files that should be registered.

		Returns:
			A list of the names of the registered config files.

		'''
		raise NotImplementedError


	def iterate_configs(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered config files.

		Returns:
			An iterator over all registered config files.

		'''
		raise NotImplementedError


	def find_local_config_entry(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the entry for the config file with the given name in the registry.

		Args:
			name: Name of the config file to find.
			default: Default value to return if the artifact is not found.

		Returns:
			The entry for the config file with the given name.

		Raises:
			:class:`KeyError`: If the config file is not registered.

		'''
		raise NotImplementedError


	def find_project_config_entry(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the entry for a config by name. Including checking the nonlocal projects.

		Args:
			name: of the config file used to register the config
			default: Default value to return if the artifact is not found.

		Returns:
			The entry for the config

		Raises:
			ConfigNotFoundError: if the config is not registered

		'''
		raise NotImplementedError


	def find_config_path(self, name: str, default: Optional[Any] = unspecified_argument) -> Path:
		'''
		Finds the path to the config file with the given name in the registry.

		Args:
			name: Name of the config file to find.
			default: Default value to return if the artifact is not found.

		Returns:
			The path to the config file with the given name.

		Raises:
			:class:`KeyError`: If the config file is not registered.

		'''
		raise NotImplementedError


	def parse_argv(self, argv: Sequence[str], script_name: Optional[str] = unspecified_argument) -> AbstractConfig:
		'''
		Parses the given arguments and returns a config object.

		Arguments are expected in the following order (all of which are optional):
			1. Meta rules to modify the config loading process and run mode.
			2. Name of the script to run.
			3. Names of registered config files that should be loaded and merged (in order of precedence).
			4. Manual config parameters (usually keys, prefixed by :code:`--` and corresponding values)

		Args:
			argv: List of arguments to parse (expected to be :code:`sys.argv[1:]`).
			script_name: Manually specified name of the script (defaults to what is specified in the resulting config).

		Returns:
			Config object containing the parsed arguments.

		'''
		raise NotImplementedError


	def create_config(self, configs: Optional[Sequence[str]] = None, data: Optional[JSONABLE] = None) -> AbstractConfig:
		'''
		Creates a config object from the given config file names and provided arguments.

		Args:
			configs: Names of registered config files to load and merge (in order of precedence).
			data: Manual config parameters to populate the config object with.

		Returns:
			Config object resulting from loading/merging `configs` and including `data`.

		'''
		raise NotImplementedError


	def load_raw_config(self, path: Union[str, Path]) -> JSONABLE:
		'''
		Loads a config file from the given path and returns the raw config data
		(made up of standard python objects, such as :py:class:`dict` and :py:class:`list`).

		Args:
			path: Path to the config file to load.

		Returns:
			Config data loaded from the given path.

		Raises:
			:class:`FileNotFoundError`: If the config file does not exist.
			:class:`ValueError`: If the config file cannot be loaded.

		'''
		raise NotImplementedError


	def configurize(self, raw: JSONABLE) -> AbstractConfig:
		'''
		Converts the given raw config data into a config object.

		Raw config data can include primitives, lists, and dicts (including OrderedDict or tuples),
		but should not include any other types.

		Args:
			raw: Raw config data to convert into a config object.

		Returns:
			Config object created from the given raw config data.

		'''
		raise NotImplementedError


	def merge_configs(self, *configs: AbstractConfig) -> AbstractConfig:
		'''
		Merges the given config objects into a single config object.

		Including recursively merging any nested config objects and overwriting any duplicate keys
		in reverse provided order (so the configs should be provided in order of precedence).

		Args:
			*configs: Provided config objects to merge.

		Returns:
			Config object resulting from merging the given config objects.

		'''
		raise NotImplementedError


	@staticmethod
	def update_config(base: AbstractConfig, update: AbstractConfig) -> AbstractConfig:
		'''
		Updates the config object with the given config object (recursively overwriting common keys).

		Args:
			base: Config object to update.
			update: Config object to update `base` with.

		Returns:
			The updated config object `base`.

		'''
		raise NotImplementedError



class AbstractCustomArtifact:
	@staticmethod
	def top(config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		raise NotImplementedError


	def get_wrapped(self) -> Union[Callable, Type]:
		raise NotImplementedError



class AbstractCreator:
	'''Abstract class for creators.'''

	_creator_name = None

	@classmethod
	def replace(cls, creator: 'AbstractCreator', config: AbstractConfig, **kwargs) -> 'AbstractCreator':
		'''
		Extracts any required information from the required `creator` to create and return a new creator.

		Args:
			creator: base creator to extract information from.
			config: config object to initialize the new creator with.
			**kwargs: other keyword arguments to pass to the new creator initialization.

		Returns:
			New creator created from the given `creator` and `config`.

		'''
		return cls(config, **kwargs)


	def __init__(self, config: AbstractConfig, **kwargs: Any):
		super().__init__(**kwargs)


	def create(self, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Creates an object from the given config node and other arguments.

		Args:
			config: Config node to create the object from.
			*args: Manual arguments to pass to the object.
			**kwargs: Manual keyword arguments to pass to the object.

		Returns:
			Object created from the given config node and other arguments.

		'''
		return self.create_product(config, args=args, kwargs=kwargs)


	def create_product(self, config: AbstractConfig, args: Optional[Tuple] = None,
	                   kwargs: Optional[Dict[str,Any]] = None) -> Any:
		'''
		Creates an object from the given config node and other arguments.

		Args:
			config: Config node to create the object from.
			args: Manual arguments to pass to the object.
			kwargs: Manual keyword arguments to pass to the object.

		Returns:
			Object created from the given config node and other arguments.

		'''
		raise NotImplementedError



class AbstractRunMode(Activatable):
	'''Abstract class for run modes. Run modes include Projects and Profiles'''

	def main(self, argv: Sequence[str], script_name: Optional[str] = None) -> Any:
		'''
		Runs the script with the given arguments using the config object obtained by parsing ``argv``.

		Args:
			argv: List of top-level arguments (expected to be :code:`sys.argv[1:]`).
			script_name: specified name of the script
			(defaults to what is specified in argv when it is parsed into a config object).

		Returns:
			The output of the script.

		'''
		raise NotImplementedError


	def run_script(self, script_name: str, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Runs the script with the given arguments using :func:`run()` of the current project.

		Args:
			script_name: The script name to run (must be registered).
			config: Config object to run the script with (must include the script under :code:`_meta.script_name`).
			*args: Manual arguments to pass to the script.
			**kwargs: Manual keyword arguments to pass to the script.

		Returns:
			The output of the script.

		'''
		raise NotImplementedError


	def run(self, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Runs a script with the given config object and other arguments.

		Before running the script using :func:`run_local()`, the config object is validated with :func:`validate_run()`,
		which can modify the run mode.

		Args:
			config: Config object to run the script with (must include the script under :code:`_meta.script_name`).
			args: Manual arguments to pass to the script.
			kwargs: Manual keyword arguments to pass to the script.

		Returns:
			The output of the script.

		'''
		raise NotImplementedError


	def cleanup(self):
		'''
		After running the script through :func:`main()`, this method is called
		to clean up any resources used by the run mode.

		Returns:
			:code:`None`

		'''
		for behavior in self.behaviors():
			behavior.cleanup()


	def behaviors(self) -> Iterator['AbstractBehavior']:
		'''Iterates over all behaviors associated with this project.'''
		raise NotImplementedError


	def validate_run(self, config: AbstractConfig) -> Optional['AbstractRunMode']:
		'''
		Validates the config object before running the script.

		Args:
			config: Config object to validate.

		Returns:
			:code:`None` if the config is valid, otherwise the run mode to transfer to.

		'''
		pass


	def parse_argv(self, argv: Sequence[str], *, script_name: Optional[str] = None) -> AbstractConfig:
		'''
		Parses the given arguments and returns a config object.

		Arguments are expected in the following order (all of which are optional):
			1. Meta rules to modify the config loading process and run mode.
			2. Name of the script to run.
			3. Names of registered config files that should be loaded and merged (in order of precedence).
			4. Manual config parameters (usually keys, prefixed by :code:`--` and corresponding values)

		Args:
			argv: List of arguments to parse (expected to be :code:`sys.argv[1:]`).
			script_name: Manually specified name of the script (defaults to what is specified in the resulting config).

		Returns:
			Config object containing the parsed arguments.

		'''
		raise NotImplementedError



class AbstractProject(AbstractRunMode, FileInfo, Activatable):
	'''
	Abstract class for projects. Projects track artifacts (eg. configs and components) in registries
	and manage loading configs and running scripts.
	'''
	def __init__(self, path: Optional[Union[str, Path]] = None, profile: 'AbstractProfile' = None, **kwargs):
		super().__init__(path, **kwargs)
		self._profile = profile


	def __eq__(self, other: 'AbstractProject') -> bool:
		'''
		Compares two projects based on their paths.

		Args:
			other: Another project object.

		Returns:
			:code:`True` if the projects have the same path, otherwise :code:`False`.

		'''
		return self.data['info_path'] == other.data['info_path']


	@property
	def profile(self) -> 'AbstractProfile':
		'''
		Profile object that the project is associated with.

		Returns:
			Profile object.

		'''
		return self._profile


	def nonlocal_projects(self) -> Iterator[NamedTuple]:
		'''
		Iterator over all projects that are related to this one, followed by all active base projects
		of the profile (without repeating any projects).

		Returns:
			An iterator over all projects that are related to this one, followed by all active base projects

		'''
		raise NotImplementedError


	def behaviors(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all behavior instances associated with this project.

		By default, this creates an instance for all behaviors in the profile.

		Returns:
			Iterator over meta rules in order of priority (highest -> least).

		'''
		raise NotImplementedError


	class UnknownArtifactError(Registry.NotFoundError):
		'''Raised when trying to find an artifact that is not registered.'''
		def __init__(self, artifact_type: str, ident: str):
			super().__init__(f'{artifact_type} {ident!r} not found')
			self.artifact_type = artifact_type
			self.ident = ident


	def xray(self, artifact: str, *, sort: Optional[bool] = False, reverse: Optional[bool] = False,
	         as_list: Optional[bool] = False) -> Optional[List[NamedTuple]]:
		'''
		Prints a list of all artifacts of the given type accessible from this project
		(including related and active base projects).

		Args:
			artifact: artifact type (e.g. 'script', 'config')
			sort: sort the list of artifacts by name
			reverse: reverse the order of the list of artifacts
			as_list: instead of printing, return the list of artifacts

		Returns:
			list: if as_list is True, returns a list of artifacts

		Raises:
			UnknownArtifactTypeError: if the given artifact type does not exist

		'''
		raise NotImplementedError


	def find_local_artifact(self, artifact_type: str, ident: str,
	                        default: Optional[Any] = unspecified_argument) -> Optional[NamedTuple]:
		'''
		Finds the artifact with the given type and identifier in ``self``
		without checking related projects or active projects in the profile.

		Args:
			artifact_type: The type of artifact to find.
			ident: The identifier of the artifact to find.
			default: The default value to return if the artifact is not found.

		Returns:
			The artifact entry from the registry corresponding to the given type.

		Raises:
			UnknownArtifactTypeError: If the artifact type is not registered.
			UnknownArtifactError: If the artifact is not found and no default is given.

			'''
		raise self.UnknownArtifactError(artifact_type, ident)


	def find_artifact(self, artifact_type: str, ident: str,
	                  default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds an artifact in the project's registries.
		Artifacts are data or functionality such as configs and components.

		Args:
			artifact_type: Type of artifact to find (eg. 'config' or 'component').
			ident: Name of the artifact that was registered.
			default: Default value to return if the artifact is not found.

		Returns:
			Artifact object, or, if a default value is given and artifact is not found.

		Raises:
			:class:`UnknownArtifactError`: if the artifact is not found and no default is specified.

		'''
		raise self.UnknownArtifactError(artifact_type, ident)


	def register_artifact(self, artifact_type: str, ident: str, artifact: Union[Type, Callable],
	                      **kwargs: Any) -> NamedTuple:
		'''
		Registers a new artifact in the project's registries.
		Overwrites any existing artifact with the same name.

		Args:
			artifact_type: Type of artifact to register (eg. 'config' or 'component').
			ident: Name of the artifact to be registered.
			artifact: Artifact object to register (usually a function, type, or path).
			**kwargs: Optional additional parameters to store with the artifact.

		Returns:
			Registration entry for the artifact.

		'''
		raise NotImplementedError


	def iterate_artifacts(self, artifact_type: str) -> Iterator[NamedTuple]:
		'''
		Iterates over all artifacts of the given type in the project's registries.

		Args:
			artifact_type: Type of artifact to iterate (eg. 'config' or 'component').

		Returns:
			Iterator over all artifacts of the given type.

		'''
		raise NotImplementedError


	def create_config(self, *configs: str, **parameters: Primitive) -> AbstractConfig:
		'''
		Creates a new config object with the given parameters.Creates a config object from
		the given config file names and provided parameters.

		Args:
			*configs: Names of registered config files to load and merge (in order of precedence).
			**parameters: Manual config parameters to populate the config object with.

		Returns:
			Config object.

		'''
		raise NotImplementedError


	def quick_run(self, script_name: str, *configs: str, **parameters: JSONABLE) -> Any:
		'''
		Composes :func:`create_config() and :func:`run_script()` into a single method to
		first create a config object and then run the specified script.

		Args:
			script_name: Name of the script to run (should be registered).
			*configs: Names of registered config files to load and merge (in order of precedence).
			**parameters: Manual config parameters to populate the config object with.

		Returns:
			Output of the script.

		'''
		config = self.create_config(*configs, **parameters)
		return self.run_script(script_name, config)



class AbstractProfile(FileInfo, Activatable): # generally you should extend organization.workspaces.ProfileBase
	'''
	Abstract classes for profiles. Profiles manage projects and provide meta rules. Unlike projects,
	generally a runtime should only use a single global instance of a profile.
	'''

	@classmethod
	def get_project_type(cls, ident: str) -> NamedTuple:
		'''
		Gets the project type entry for the given identifier (from a registry).

		Args:
			ident: Name of the registered project type.

		Returns:
			Project type entry.

		'''
		raise NotImplementedError


	@classmethod
	def replace_profile(cls, profile: 'AbstractProfile') -> 'AbstractProfile':
		'''
		Replaces the current profile instance with the given profile. This is used to set the global profile.

		Args:
			profile: New profile instance.

		Returns:
			Old profile instance (which is now replaced).

		'''
		raise NotImplementedError


	@classmethod
	def get_profile(cls) -> 'AbstractProfile':
		'''
		Gets the current profile instance of the runtime environment.

		Returns:
			Profile instance.

		'''
		raise NotImplementedError


	@classmethod
	def register_behavior(cls, name: str, typ: Type['AbstractBehavior'], *,
	                      description: Optional[str] = None) -> NamedTuple:
		'''
		Registers a new behavior in the profile.

		Behaviors are classes which are instantiated and managed by .

		Args:
			name: Name of the behavior.
			typ: Behavior class (recommended to subclass :class:`AbstractBehavior`).
			description: Description of the behavior.

		Returns:
			Registration entry for the behavior.

		'''
		raise NotImplementedError


	@classmethod
	def get_behavior(cls, name: str) -> NamedTuple:
		'''
		Gets the behavior entry for the given identifier (from the registry).

		Args:
			name: Name of the registered behavior.

		Returns:
			Behavior entry.

		'''
		raise NotImplementedError


	@classmethod
	def iterate_behaviors(cls) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered behaviors.

		Returns:
			Iterator over all behavior entries.

		'''
		raise NotImplementedError


	def entry(self, script_name: Optional[str] = None) -> None:
		'''
		Primary entry point for the profile. This method is called when using the :command:`fig` command.

		Args:
			script_name: Manually specified script name to run (if not provided, will be parsed from :code:`sys.argv`).

		Returns:
			:code:`None`

		'''
		raise NotImplementedError


	def initialize(self, *projects: str) -> None:
		'''
		Initializes the specified projects (including activating them, which generally registers
		all associated configs and imports files and packages)

		Args:
			*projects: Identifiers of projects to initialize (activates the current project only, if none is provided).

		Returns:
			:code:`None`

		'''
		raise NotImplementedError


	def main(self, argv: Sequence[str], *, script_name: Optional[str] = None) -> Any:
		'''
		Runs the script with the given arguments using :func:`main()` of the current project.

		Args:
			argv: List of top-level arguments (expected to be :code:`sys.argv[1:]`).
			script_name: specified name of the script
			(defaults to what is specified in argv when it is parsed into a config object).

		Returns:
			The output of the script.

		'''
		raise NotImplementedError


	def run_script(self, script_name: str, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Runs the script registered with the given name and the given arguments using
		:func:`run_script()` of the current project.

		Args:
			script_name: Name of the script to run (must be registered).
			config: Config object to run the script with.
			*args: Manual arguments to pass to the script.
			**kwargs: Manual keyword arguments to pass to the script.

		Returns:
			The output of the script.

		'''
		raise NotImplementedError


	def run(self, config: AbstractConfig, *,
	        args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None) -> Any:
		'''
		Runs the script with the given arguments using :func:`run()` of the current project.

		This method assumes the script_name is already contained in the config, otherwise use :func:`run_script()`.

		Args:
			config: Config object to run the script with (must include the script under :code:`_meta.script_name`).
			args: Manual arguments to pass to the script.
			kwargs: Manual keyword arguments to pass to the script.

		Returns:
			The output of the script.

		'''
		raise NotImplementedError


	def quick_run(self, script_name: str, *configs: str, **parameters: Any) -> Any:
		'''
		Creates a config object and runs the script using :func:`quick_run()` of the current project.

		Args:
			script_name: Name of the script to run (must be registered).
			*configs: Names of registered config files to load and merge (in order of precedence).
			**parameters: Manual config parameters to populate the config object with.

		Returns:
			Output of the script.

		'''
		raise NotImplementedError


	def cleanup(self) -> None:
		'''
		Calls :func:`cleanup()` of the current project. Generally not needed to be called manually.

		Returns:
			:code:`None`

		'''
		raise NotImplementedError


	def create_config(self, *configs: str, **parameters: JSONABLE) -> AbstractConfig:
		'''
		Process the provided data to create a config object (using the current project).

		Args:
			configs: usually a list of parent configs to be merged
			parameters: any manual parameters to include in the config object

		Returns:
			Config object resulting from loading/merging `configs` and including `data`.
		'''
		raise NotImplementedError


	def parse_argv(self, argv: Sequence[str], script_name=None) -> AbstractConfig:
		'''
		Parses the given arguments and returns a config object.

		Arguments are expected in the following order (all of which are optional):
			1. Meta rules to modify the config loading process and run mode.
			2. Name of the script to run.
			3. Names of registered config files that should be loaded and merged (in order of precedence).
			4. Manual config parameters (usually keys, prefixed by :code:`--` and corresponding values)

		Args:
			argv: List of arguments to parse (expected to be :code:`sys.argv[1:]`).
			script_name: Manually specified name of the script (defaults to what is specified in the resulting config).

		Returns:
			Config object containing the parsed arguments.

		'''
		raise NotImplementedError


	def extract_info(self, other: 'AbstractProfile') -> None:
		'''
		Extract data from the provided profile instance and store it in self.

		Recommended to use if a project expects a custom profile different from the currently used one.

		Args:
			profile: Base profile instance.

		Returns:
			:code:`None`

		'''
		return super().extract_info(other)


	def get_current_project(self) -> AbstractProject:
		'''
		Gets the current project instance.

		Returns:
			Current project instance.

		'''
		raise NotImplementedError


	def switch_project(self, ident: Optional[str] = None) -> AbstractProject:
		'''
		Switches the current project to the one with the given identifier.

		Args:
			ident: Name of the project to switch to, defaults to the default project (with name: None).

		Returns:
			New current project instance.

		'''
		raise NotImplementedError


	def project_context(self, ident: Optional[Union[str, AbstractProject]] = None) -> ContextManager[AbstractProject]:
		'''
		Context manager to temporarily switch to a different current project.

		Args:
			ident: Name of the project to switch to, defaults to the default project (with name: None).

		Returns:
			Context manager to switch to the specified project.

		'''
		raise NotImplementedError


	def iterate_projects(self) -> Iterator[AbstractProject]:
		'''
		Iterates over all loaded projects.

		Returns:
			Iterator over all loaded project instances.

		'''
		raise NotImplementedError


	def get_project(self, ident: Optional[str] = None) -> AbstractProject:
		'''
		Gets the project with the given identifier, if the project is not already loaded, it will be loaded.

		Args:
			ident: Name of the project to get, defaults to the default project (with name: None).

		Returns:
			Project instance.

		'''
		raise NotImplementedError



class AbstractBehavior:
	'''Interface for meta rules.'''

	def __init__(self, project: AbstractProject, **kwargs: Any):
		'''
		Behaviors are usually instantiated by the project at the beginning :meth:`main()` method.

		Otherwise, if :meth:`main()` is not used, the behaviors are instantiated in :meth:`run()`
		right before the behaviors are filtered using :meth:`include()`.

		Args:
			project: Project of this behavior instance.
			**kwargs: Additional keyword arguments (unused).

		'''
		super().__init__(**kwargs)


	class TerminationFlag(KeyboardInterrupt): # TODO: can decide the output
		'''Raised if the subsequent script should not be run.'''

		out = None
		def __init__(self, out: Any = None):
			'''
			Prevents the subsequent script from being run.

			Args:
				out: User-defined output to be returned instead of the output of the script.
			'''
			super().__init__()
			self.out = out


	@staticmethod
	def parse_argv(meta: Dict[str, Any], argv: List[str], script_name: Optional[str] = None) -> Optional[List[str]]:
		'''
		Optionally modifies the arguments when the project's :meth:`main()` is called.

		Args:
			meta: Meta-data extracted from the argv so far (can be modified here).
			argv: List of arguments to parse (expected to be :code:`sys.argv[1:]`).
			script_name: Manually specified name of the script (if not provided, it will be parsed from argv).

		Returns:
			Modified list of arguments (or :code:`None` if no modification is needed).

		'''
		pass

	@staticmethod
	def validate_project(config: AbstractConfig) -> Optional[AbstractProject]:
		'''
		Validates the project (provided in constructor) using the given config object before running it.

		If a different project should be used, it can be returned here.

		Args:
			config: The config object to use for validation.

		Returns:
			The new project to use for the script execution or ``None`` if no new project was returned.

		'''
		pass


	@staticmethod
	def include(meta: AbstractConfig) -> bool:
		'''
		Checks if the current behavior should be included in the subsequent script execution.

		Args:
			meta: Meta config to configure behavior and script execution.

		Returns:
			:code:`True` if the behavior should be included, :code:`False` otherwise.

		'''
		return True


	@staticmethod
	def pre_run(meta: AbstractConfig, config: AbstractConfig) -> Optional[AbstractConfig]:
		'''
		Runs before the script is executed, within the project :meth:`run()`.

		Args:
			meta: Meta config to configure behavior and script execution.
			config: Config object which will be passed to the script.

		Returns:
			New config object to pass to the script instead (or None to use the original config).

		Raises:
			:exc:`TerminationFlag`: If the subsequent script should not be run.

		'''
		pass


	class IgnoreException(Exception):
		'''Raised if the underlying exception raised while running the script should be ignored.'''
		def __init__(self, out: Any = None):
			'''
			Instead of raising the exception, ``out`` will be returned.

			Args:
				out: User-defined output to be returned instead of raising the exception.
			'''
			super().__init__()
			self.out = out


	@staticmethod
	def handle_exception(meta: AbstractConfig, config: AbstractConfig, exc: Exception) -> None:
		'''
		Runs if the script raises an exception.

		Args:
			meta: Meta config to configure behavior and script execution.
			config: Config object to run the script with.
			exc: Exception that was raised.

		Returns:
			:code:`None`

		Raises:
			:exc:`IgnoreException`: If the exception should be ignored.
			:exc:`TerminationFlag`: If the subsequent behaviors should not be run.

		'''
		pass


	@staticmethod
	def post_run(meta: AbstractConfig, config: AbstractConfig, output: Any) -> Optional[Any]:
		'''
		Function called after the script is run.
		If the function returns a value, it will be used as the output of the script.

		Args:
			meta: Meta config to configure behavior and script execution.
			config: Config object used to run the script with.
			output: Output of the script.

		Returns:
			New output of the script, or :code:`None` to use the original output.

		Raises:
			:exc:`TerminationFlag`: If the subsequent behaviors should not be run.

		'''
		pass


	@staticmethod
	def cleanup():
		'''
		Function called at the end of the project :meth:`main()` during cleanup.

		Note that this is only called if the project :meth:`main()`. To define a cleanup function for
		the behavior that is called every time a script is run, use :meth:`post_run()`.

		Returns:
			:code:`None`

		'''
		pass


	def __cmp__(self, other):
		'''Compares the behavior to another behavior for ordering.'''
		raise NotImplementedError




import sys
import io
from contextlib import redirect_stdout

import omnifig as fig

from _test_util import reset_profile

# def test_debug():
# 	pass
#
# def test_help():
# 	pass
#



def test_help():
	reset_profile()
	proj = fig.switch_project('example6')
	proj.activate()

	argv = ['-h', '--y', '99']

	print()
	out = fig.main(argv)

	assert out is None



def test_help2():
	reset_profile()
	proj = fig.switch_project('example6')
	proj.activate()

	argv = ['add', '-h', '--y', '99']

	print()
	out = fig.main(argv)

	assert out is None


def test_help3():
	reset_profile()
	fig.switch_project('example6')

	argv = ['repeater', '-h', '--y', '99']

	print()
	out = fig.main(argv)

	assert out is None



def test_help4():
	reset_profile()
	fig.switch_project('example6')

	argv = ['add', '-h', '--y', '99', '--_meta.help', 'no'] # you cant override the behavior manually

	print()
	out = fig.main(argv)

	assert out is None



def test_help5():
	reset_profile()
	fig.switch_project('example6')

	argv = ['-h', '--y', '99', '--verbose']

	print()
	out = fig.main(argv)

	assert out is None




def test_help6():
	reset_profile()
	fig.switch_project('example6')

	argv = ['-h', '--y', '99', '--verbose']

	print()
	out = fig.main(argv, 'is-quiet')

	assert out is None



def test_help7():
	reset_profile()
	proj = fig.switch_project('example6')

	argv = ['fig', 'mirror', '-h']

	print()
	old = sys.argv
	sys.argv = argv
	out = fig.entry()
	sys.argv = old

	assert out is None



def test_help8():
	reset_profile()
	proj = fig.switch_project('example6')

	argv = ['--y', '99', '-h']

	print()
	out = fig.main(argv)

	assert out is None




def test_debug():
	reset_profile()
	proj = fig.switch_project('example6')
	proj.activate()

	print()

	config = fig.create_config('c2', **{'_meta.debug': True})

	assert config.cro == ('c2', 'the-kids/kendall')
	assert config.bases == ('c2',)

	assert config.pull('_meta.debug', silent=True) is True
	assert config.pull('_meta.quiet', False, silent=True) is False

	out = fig.run_script('add', config)

	assert out == 3
	assert config.pull('_meta.quiet', False, silent=True) is True

	out = fig.run_script('add', config, 2, 2)

	assert out == 5 # a bit orwellian

	assert config.cro == ('debug', 'c1', 'c2', 'the-kids/kendall')
	assert config.bases == ('debug',)

	config.push('x', 3)
	out = fig.run_script('add', config, y=2)

	assert out == 5



def test_debug2():
	reset_profile()
	fig.switch_project('example5')

	print()
	out = fig.main(['add-1', '-d', '--a=99'])

	assert out == 100


def test_debug3():
	reset_profile()
	fig.switch_project('example6')

	argv = ['repeater', 'c2', '-d', '--y=99']

	out = fig.main(argv)

	assert out == (3, 11)

	old = sys.argv
	sys.argv = argv
	out = fig.entry('add')
	sys.argv = old

	assert out is None

	argv = ['repeater', 'c2', '--y=99']

	out = fig.main(argv)

	assert out == (104, 15)



def test_debug4():

	reset_profile()
	fig.switch_project('example6')

	out = fig.quick_run('repeater', 'c2', y=99, **{'_meta.debug': True})

	assert out == (3, 11)



def test_quiet():
	reset_profile()

	fig.initialize('example6')

	config = fig.parse_argv(['-q', '--b', 'baba', '--x=99'])

	assert not config.silent

	buf = io.StringIO()

	with redirect_stdout(buf):
		x, y, args, kwargs = fig.run_script('mirror', config, 'arg', z=100)

	assert buf.getvalue() == ''

	assert not config.silent
	assert x == 99
	assert y == 2
	assert args == ('arg',)
	assert kwargs == {'z': 100}



def test_quiet2():
	fig.initialize('example6')

	config = fig.parse_argv(['--b', 'baba', '--x=99'])

	assert not config.silent
	buf = io.StringIO()

	with redirect_stdout(buf):
		fig.run_script('mirror', config, 'arg', z=100)

	assert not config.silent
	assert buf.getvalue() != ''



def test_quiet3():
	fig.initialize('example6')

	config = fig.parse_argv(['-q', '--b', 'baba', '--x=99'])

	assert not config.silent
	config.push('_meta.quiet', False)

	buf = io.StringIO()

	with redirect_stdout(buf):
		fig.run_script('mirror', config, 'arg', z=100)

	assert not config.silent
	assert buf.getvalue() != ''



def test_quiet4():
	reset_profile()

	fig.initialize('example6')

	config = fig.parse_argv(['--b', 'baba', '--x=99'])

	config.silent = True
	assert config.silent

	buf = io.StringIO()

	with redirect_stdout(buf):
		x, y, args, kwargs = fig.run_script('mirror', config, 'arg', z=100)

	assert config.silent
	assert buf.getvalue() == ''



def test_quiet5():
	reset_profile()

	fig.initialize('example6')

	config = fig.parse_argv(['-q', '--b', 'baba', '--x=99'])

	config.silent = True
	assert config.silent

	buf = io.StringIO()

	with redirect_stdout(buf):
		x, y, args, kwargs = fig.run_script('mirror', config, 'arg', z=100)

	assert config.silent
	assert buf.getvalue() == ''



def test_quiet6():
	reset_profile()

	fig.initialize('example6')

	config = fig.parse_argv(['-q', '--b', 'baba', '--x=99'])

	assert fig.run_script('is-quiet', config) is True



def test_quiet7():
	reset_profile()

	fig.initialize('example6')

	config = fig.parse_argv(['--b', 'baba', '--x=99', '--_meta.quiet', 'yes'])

	assert fig.run_script('is-quiet', config) is True






from typing import Dict, Union, Any
import yaml
from pathlib import Path
from collections import OrderedDict
from omnibelt import load_yaml

from . import __logger__ as prt



class Activatable:
	'''
	Mix-in class for objects that can be activated and deactivated.

	Once activated, the object is in a usable state, and cannot be activated again
	(and the object can only be deactivated if it is already activated).

	Primarily used by :class:`omnifig.abstract.AbstractProfile` and :class:`omnifig.abstract.AbstractProject`.

	'''
	def __init__(self, *args: Any, **kwargs: Any):
		super().__init__(*args, **kwargs)
		self._is_activated = False


	@property
	def is_activated(self) -> bool:
		'''Flag whether the object is currently activated'''
		return self._is_activated


	def activate(self, *args: Any, **kwargs: Any) -> None:
		'''
		Top-level method to activate the object.

		Args:
			*args: Arguments to pass to the :func:`_activate` method
			**kwargs: Keyword arguments to pass to the :func:`_activate` method

		Returns:
			:code:`None`

		'''
		if self._is_activated:
			return
		self._is_activated = True
		self._activate(*args, **kwargs)


	def _activate(self, *args: Any, **kwargs: Any) -> None:
		'''
		Internal method to activate the object. This should be overridden by subclasses.

		This method will only be called once, and only if the object is not already activated.

		Args:
			*args: Any arguments passed to the :func:`activate()` method
			**kwargs: Keyword arguments passed to the :func:`activate()` method

		Returns:
			:code:`None`

		'''
		pass


	def deactivate(self, *args, **kwargs):
		'''
		Top-level method to deactivate the object.

		Args:
			*args: Arguments to pass to the :func:`_deactivate()` method
			**kwargs: Keyword arguments to pass to the :func:`_deactivate()` method

		Returns:
			:code:`None`

		'''
		if not self._is_activated:
			return
		self._deactivate(*args, **kwargs)
		self._is_activated = False


	def _deactivate(self, *args, **kwargs):
		'''
		Internal method to deactivate the object. This should be overridden by subclasses.

		This method will only be called once, and only if the object is currently activated.

		Args:
			*args: Any arguments passed to the :func:`deactivate()` method
			**kwargs: Keyword arguments passed to the :func:`deactivate()` method

		Returns:
			:code:`None`

		'''
		pass



class FileInfo:
	'''Mix-in class for objects that loads and stores information from a file.'''


	@staticmethod
	def load_raw_info(path: Path) -> Dict[str, Any]:
		'''
		Loads the info yaml file at the given path.

		Args:
			path: File path containing the info yaml file

		Returns:
			:code:`dict` containing the loaded info

		'''
		raw = None
		try:
			if path.exists():
				raw = load_yaml(path, ordered=True)
		except yaml.YAMLError:
			prt.error(f'Error loading yaml file: {path}')
		if raw is None:
			raw = {}
		raw['info_path'] = str(path) # automatically set info_path to the path
		raw['info_dir'] = str(path.parent)
		return raw


	def __init__(self, data: Union[str, Path, Dict[str, Any]] = None, **kwargs: Any):
		'''
		Loads the info if the provided `data` is a file path.

		Args:
			data: A file path to a yaml file, or a dictionary containing the info
			**kwargs: Other arguments passed on to ``super()``
		'''
		super().__init__(**kwargs)
		if isinstance(data, str):
			data = Path(data)
		if isinstance(data, Path):
			data = self.load_raw_info(data)
		if data is None:
			data = OrderedDict()
		self.data = data


	@property
	def name(self):
		'''The name of the object, as specified in the info file.'''
		return self.data.get('name', '')


	def __repr__(self):
		return f'{self.__class__.__name__}({self.name})'


	def __str__(self):
		return f'{self.__class__.__name__}[{self.name}]({", ".join(self.data.keys())})'


	def extract_info(self, other: 'FileInfo'):
		'''
		Extracts the info from the given object and stores it in this object.

		Usually used to replace :code:`other` with :code:`self`.

		Args:
			other: The source object to extract the info from

		Returns:
			:code:`None`

		'''
		self.data = other.data

















from functools import reduce
import random

import omnifig as fig

class Toy_Dataset(fig.Configurable):
	def __init__(self, dataroot, input_dim, output_dim, name='unknown dataset',
	             labeled=True, num_categories=None, num_samples=1000, **kwargs):

		# input_dim = A.pull('input_dim', '<>input_shape', silent=True)
		if isinstance(input_dim, (list, tuple)):
			input_dim = reduce((lambda x, y: x * y), input_dim)

		# output_dim = A.pull('output_dim', '<>output_shape', silent=True)
		if isinstance(output_dim, (list, tuple)):
			output_dim = reduce((lambda x, y: x * y), output_dim)

		if num_categories is None:
			num_categories = output_dim

		super().__init__(**kwargs)

		self.dataroot = dataroot
		self.name = name

		self.labeled = labeled
		self.num_categories = num_categories
		
		self.input_dim = input_dim
		self.output_dim = output_dim if output_dim is not None and self.labeled else input_dim
		
		self.samples = list(range(num_samples))
		self.idx = 0
		
	def get_io(self):
		return self.input_dim, self.output_dim
		
	def __len__(self):
		return len(self.samples)
		
	def __getitem__(self, idx):
		val = self.samples[idx]
		return val**0.8, val % self.num_categories # doesn't look much like MNIST or CIFAR :)
		
	def __next__(self):
		if self.idx == len(self):
			raise StopIteration
		
		x,y = self[self.idx]
		
		self.idx += 1
		
		return x,y
		
	def __iter__(self):
		self.idx = 0
		return self


@fig.component('mnist')
class MNIST(Toy_Dataset):
	def __init__(self, name='mnist', input_dim=(28, 28, 1), output_dim=10, **kwargs):
		super().__init__(name=name, input_dim=input_dim, output_dim=output_dim, **kwargs)


@fig.component('cifar')
class CIFAR(Toy_Dataset):
	def __init__(self, name='cifar', in_color=True, input_dim=None, output_dim=None, num_categories=10, **kwargs):
		if input_dim is None:
			input_dim = (32, 32, 3) if in_color else (32, 32, 1)
		if output_dim is None:
			output_dim = num_categories
		assert num_categories in {10, 100}, f'CIFAR image dataset has either 10 or 100 categories, ' \
		                                    f'not {num_categories}'
		super().__init__(name=name, input_dim=input_dim, output_dim=output_dim, num_categories=num_categories, **kwargs)
		
		
	def _make_grayscale(self):
		pass # do something
	
	
@fig.modifier('even')
class Even(Toy_Dataset):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.samples = [x for x in self.samples if x % 2 == 0]


@fig.modifier('noisy')
class Noisy(Toy_Dataset):
	def __init__(self, *args, noise_weight=None, **kwargs):
		super().__init__(*args, **kwargs)
		self.noise_weight = noise_weight

	def __next__(self):
		x, y = super().__next__()
		
		if self.noise_weight is not None and self.noise_weight > 0:
			x += random.gauss(0,self.noise_weight)

		return x, y


@fig.modifier('rotate')
class Rotated(Toy_Dataset):
	@fig.config_aliases(rotation_prob='prob')
	def __init__(self, *args, rotation_prob=0.5, **kwargs):
		super().__init__(*args, **kwargs)
		self.rotation_prob = rotation_prob

	def __next__(self):
		x, y = super().__next__()
		if random.random() < self.rotation_prob:
			x = x * 1j
		return x, y


@fig.modifier('shuffled')
class Shuffled(Toy_Dataset):
	def __init__(self, *args, shuffle=True, **kwargs):
		super().__init__(*args, **kwargs)
		if shuffle:
			random.shuffle(self.samples)


@fig.modifier('subset')
class Subset(Toy_Dataset):
	def __init__(self, *args, limit=None, **kwargs):
		super().__init__(*args, **kwargs)
		if limit < len(self):
			self.samples = self.samples[:limit]









from typing import Dict, Optional, Union, Any, Iterator, NamedTuple, Type, List
import os
from itertools import chain
from collections import deque
from pathlib import Path
from tabulate import tabulate
from omnibelt import unspecified_argument, Class_Registry, colorize

from .. import __logger__ as prt
from ..abstract import AbstractProject, AbstractCreator

from .profiles import ProfileBase
from .workspaces import ProjectBase, GeneralProject


class Project(GeneralProject, name='default'):
	class Creator_Registry(Class_Registry, components=['project', 'description']):
		'''Registry for creators which determine how components are instantiated from the config.'''
		pass
	
	class Component_Registry(Class_Registry, components=['creator', 'project', 'description']):
		'''Registry for components (classes) which can be instantiated through the config.'''
		pass
	
	class Modifier_Registry(Class_Registry, components=['project', 'description']):
		'''
		Registry for modifiers (classes) which can modify components through the config
		by dynamically defining subclasses.
		'''
		pass
	
	def __init__(self, path: Optional[Union[str, Path]] = None, *,
	             creator_registry: Optional[Creator_Registry] = None,
	             component_registry: Optional[Component_Registry] = None,
	             modifier_registry: Optional[Modifier_Registry] = None,
	             **kwargs):
		if path is None:
			path = Path().cwd()
		if creator_registry is None:
			creator_registry = self.Creator_Registry()
		if component_registry is None:
			component_registry = self.Component_Registry()
		if modifier_registry is None:
			modifier_registry = self.Modifier_Registry()
		super().__init__(path, **kwargs)
		self._artifact_registries.update({
			'creator': creator_registry,
			'component': component_registry,
			'modifier': modifier_registry,
		})
	
	def related(self) -> Iterator[AbstractProject]:
		'''
		Iterate over all projects related to this one (based on ``related`` in the project info file).

		Returns:
			An iterator over all projects related to this one.

		'''
		for ident in self.data.get('related', []):
			try:
				yield self._profile.get_project(ident)
			except self._profile.UnknownProjectError:
				pass
	
	def nonlocal_projects(self) -> Iterator[NamedTuple]:
		'''
		Iterator over all projects that are related to this one, followed by all active base projects
		of the profile (without repeating any projects).

		Returns:
			An iterator over all projects that are related to this one, followed by all active base projects

		'''
		past = {self}
		for proj in chain(self.related(), self._profile.iterate_base_projects()):
			if proj in past:
				continue
			yield proj
			past.add(proj)
	
	def xray(self, artifact: str, *, include_nonlocal: Optional[bool] = True, as_list: Optional[bool] = False,
	         sort: Optional[bool] = False, reverse: Optional[bool] = False) -> Optional[List[NamedTuple]]:
		'''
		Prints a list of all artifacts of the given type accessible from this project
		(including related and active base projects).

		Args:
			artifact: artifact type (e.g. 'script', 'config')
			include_nonlocal: whether to include artifacts from non-local projects
				(related and active base projects)
			as_list: instead of printing, return the list of artifacts
			sort: sort the list of artifacts by name
			reverse: reverse the order of the list of artifacts

		Returns:
			list: if as_list is True, returns a list of artifacts

		Raises:
			UnknownArtifactTypeError: if the given artifact type does not exist

		'''
		if not include_nonlocal:
			return super().xray(artifact, as_list=as_list, sort=sort, reverse=reverse)
		
		local = super().xray(artifact, reverse=False, as_list=True, sort=sort)
		
		vocab = {e.name for e in local}
		full = [(entry, False) for entry in local]
		
		for proj in self.nonlocal_projects():
			for entry in proj.xray(artifact, include_nonlocal=False, reverse=False, as_list=True, sort=sort):
				name = entry.name
				full.append((entry, name in vocab))
				vocab.add(name)
		
		if sort:
			full.sort(key=lambda e: e[0].name)
		if reverse:
			full.reverse()
		
		if as_list:
			return [e[0] for e in full]
		
		if len(full):
			rows, descs = zip(*[self._format_xray_entry(t) for t, _ in full])
			for row, (entry, is_dup) in zip(rows, full):
				if is_dup:
					row[0] = colorize(row[0], color="red")
			descs = [None if d is None else f'\t[{d}]' for d in descs]
			
			table = tabulate(rows, tablefmt='simple')
			
			lines = [line for lines in zip(table.splitlines()[1:-1], descs) for line in lines if line is not None]
			print('\n'.join(lines))
	
	def missing_related(self) -> Iterator[str]:
		'''
		Iterate over all projects related to this one that cannot be found by the current profile.

		Returns:
			An iterator over all projects related to this one that cannot be found by the current profile.

		'''
		for ident in self.data.get('related', []):
			try:
				self._profile.get_project(ident)
			except self._profile.UnknownProjectError:
				yield ident
	
	# region Registration
	def find_creator(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the creator with the given name.

		Args:
			name: the creator was registered with.
			default: default value to return if the creator is not found.

		Returns:
			The creator entry corresponding to the given name.

		Raises:
			UnknownArtifactError: If the creator is not found and no default is given.

		'''
		return self.find_artifact('creator', name, default=default)
	
	def register_creator(self, name: str, typ: Type[AbstractCreator], *,
	                     description: Optional[str] = None) -> NamedTuple:
		'''
		Register a creator with the given name.

		Args:
			name: to register the script under
			typ: the creator type (should be a subclass of :class:`AbstractCreator`)
			description: description of the creator

		Returns:
			The entry of the creator that was registered

		'''
		return self.register_artifact('creator', name, typ, description=description)
	
	def iterate_creators(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered creator entries.

		Returns:
			An iterator over all registered creator entries.

		'''
		return self.iterate_artifacts('creator')
	
	def find_component(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the component with the given name.

		Args:
			name: the component was registered with.
			default: default value to return if the component is not found.

		Returns:
			The component entry corresponding to the given name.

		Raises:
			UnknownArtifactError: If the component is not found and no default is given.

		'''
		return self.find_artifact('component', name, default=default)
	
	def register_component(self, name: str, typ: Type, *, creator: Union[str, AbstractCreator] = None,
	                       description: Optional[str] = None) -> NamedTuple:
		'''
		Register a component with the given name.

		Args:
			name: to register the component under
			typ: the component type (recommended to be a subclass of :class:`Configurable`)
			creator: the creator to use for this component (if none is specified in the config)
			description: description of the component

		Returns:
			The entry of the component that was registered

		'''
		return self.register_artifact('component', name, typ, creator=creator, description=description)
	
	def iterate_components(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered component entries.

		Returns:
			An iterator over all registered component entries.

		'''
		return self.iterate_artifacts('component')
	
	def find_modifier(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the modifier with the given name.

		Args:
			name: the modifier was registered with.
			default: default value to return if the modifier is not found.

		Returns:
			The modifier entry corresponding to the given name.

		Raises:
			UnknownArtifactError: If the modifier is not found and no default is given.

		'''
		return self.find_artifact('modifier', name, default=default)
	
	def register_modifier(self, name: str, typ: Type, *, description: Optional[str] = None) -> NamedTuple:
		'''
		Register a modifier with the given name.

		Args:
			name: to register the modifier under
			typ: the modifier type (recommended to be a subclass of :class:`Configurable`)
			description: description of the modifier

		Returns:
			The entry of the modifier that was registered

		'''
		return self.register_artifact('modifier', name, typ, description=description)
	
	def iterate_modifiers(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered modifiers entries.

		Returns:
			An iterator over all registered modifier entries.

		'''
		return self.iterate_artifacts('modifier')
	
	# endregion
	
	def _find_nonlocal_artifact(self, artifact_type: str, ident: str) -> Optional[NamedTuple]:
		'''
		Finds the artifact with the given type and identifier in the related projects and,
		if that fails, then the active projects in the profile. Note that this starts by
		searching through the graph related projects in breadth-first order, followed by
		any active base projects in the profile (that haven't already been checked).

		Args:
			artifact_type: The type of artifact to find.
			ident: The identifier of the artifact to find.

		Returns:
			The artifact entry from the registry corresponding to the given type.

		Raises:
			UnknownArtifactTypeError: If the artifact type is not registered.
			UnknownArtifactError: If the artifact is not found and no default is given.

		'''
		past = {self}
		new = deque(self.related())
		while len(new):
			proj = new.popleft()
			past.add(proj)
			try:
				if isinstance(proj, Profile.Project):
					new.extend(p for p in proj.related() if p not in past)
					return proj.find_local_artifact(artifact_type, ident)
				return proj.find_artifact(artifact_type, ident)
			except self.UnknownArtifactError:
				pass
		
		for proj in self._profile.iterate_base_projects():
			if proj not in past:
				try:
					if isinstance(proj, Profile.Project):
						new.extend(p for p in proj.related() if p not in past)
						return proj.find_local_artifact(artifact_type, ident)
					return proj.find_artifact(artifact_type, ident)
				except self.UnknownArtifactError:
					pass
		
		raise self.UnknownArtifactError(artifact_type, ident)
	
	def find_artifact(self, artifact_type: str, ident: str,
	                  default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the artifact of the given type and registered with the given identifier by searching
		first in the current project, then any related projects, and finally any active base projects
		specified by the profile. Note, that if the artifact ident may specify a project to search in
		by prefixing the ident with the project name and a colon (e.g. ``proj:ident``).

		Args:
			artifact_type: The type of artifact to find.
			ident: The identifier of the artifact to find.
			default: The default value to return if the artifact is not found.

		Returns:
			The artifact entry from the registry corresponding to the given type.

		Raises:
			UnknownArtifactTypeError: If the artifact type is not registered.
			UnknownArtifactError: If the artifact is not found and no default is given.
			UnknownProjectError: If the project specified in the artifact ident is not found.

		'''
		if ':' in ident:
			proj, ident = ident.split(':', 1)
			return self._profile.get_project(proj).find_artifact(artifact_type, ident)
		try:
			return self.find_local_artifact(artifact_type, ident)
		except self.UnknownArtifactError:
			try:
				return self._find_nonlocal_artifact(artifact_type, ident)
			except self.UnknownArtifactError:
				pass
			if default is unspecified_argument:
				raise
			return default



class Profile(ProfileBase, default_profile=True):
	Project = Project
	
	_profile_env_variable = 'OMNIFIG_PROFILE'


	def __init__(self, data: Union[str, Path, Dict, None] = None):
		if data is None:
			data = os.environ.get(self._profile_env_variable, None)
		super().__init__(data)
		self._base_projects = []


	@property
	def projects(self) -> List[Project]:
		'''Convenience property for accessing the projects in the profile. Recemmended for debugging only.'''
		return self._loaded_projects


	def _activate(self) -> None:
		'''
		Activates the profile by loading all projects and base projects specified in the profile.

		Returns:
			None

		'''
		active_projects = self.data.get('active-projects', [])
		for project in active_projects:
			proj = self.get_project(project)
			proj.activate()
			self._base_projects.append(proj)

		self._current_project_key = None


	def initialize(self, *projects: str, **kwargs: Any) -> None:
		'''
		Initializes the profile by activating it and then activating all projects specified, also adds the
		projects to the profile's active base projects.

		Args:
			*projects: The names of projects to activate and add to the active base projects.
			**kwargs: Additional keyword arguments to pass to the project initialization methods.

		Returns:
			None

		'''
		super().initialize(*projects, **kwargs)
		self._base_projects.extend(projects)


	def iterate_base_projects(self) -> Iterator[Project]:
		'''
		Iterates through the active base projects in the profile.

		The active base projects are those specified in the profile's ``active-projects`` list, and are
		expected to be fallback projects for the current project for finding artifacts.

		Returns:
			An iterator over the active base projects.

		'''
		past = set()
		for base in self._base_projects:
			proj = self.get_project(base)
			if proj not in past:
				past.add(proj)
				yield proj


	def iterate_projects(self) -> Iterator[Project]:
		'''
		Iterates through the projects in the profile in the order they were created.
		Note that this iterator will not include duplicates, even if the same project
		is loaded under multiple names.

		Returns:
			An iterator over the projects in the profile.

		'''
		past = set()
		for project in self._loaded_projects.values():
			if project not in past:
				past.add(project)
				yield project


	class UnknownProjectError(KeyError):
		'''Raised when trying to get a project with an invalid path.'''
		pass


	def _infer_project_path(self, ident: Union[str, Path, None]) -> Optional[Union[Path, str]]:
		'''
		Checks if the directory (current working directory by default) is inside a known project directory,
		and if so, returns that project.

		Args:
			ident: of the project to get (expected to be None usually).

		Returns:

		'''
		# path = ident
		if ident is None:
			ident = Path().cwd()

		contents = set(os.listdir(ident))
		for fname in self.Project.info_file_names:
			if fname in contents:
				return ident

		known = {path: name for name, path in self.data.get('projects', {}).items()}
		if str(ident) in known:
			return ident

		if len(ident.parts) > 1:
			return self._infer_project_path(ident.parent)


	def get_project(self, ident: Union[str, Path] = None, is_current: Optional[bool] = None) -> AbstractProject:
		'''
		Gets the project with the given name/path. If no name/path is given, then the current project is returned.
		If the specified project has not been initialized, then it is created using the profile's Project class
		(but it is not activated).

		If a name is given, it must be the name of a project that has already been loaded, or it must be a key in the
		dict `projects` the profile's info file where the value is the corresponding path.

		If a path is given, the path may either be a yaml file (interpretted as the project's info file)
		or a directory (in which case the project's info file is assumed to be named `.fig.project.yaml`).

		Args:
			ident: of the project to get. If not given, then the current project is returned.
			is_current: if True, then the project is set as the current project.

		Returns:
			The project with the given name/path.

		Raises:
			UnknownProjectError: If the project is not found.
			ValueError: If the project is found, but it has a name that a different project is already using.

		'''
		# if is_current is None:
		# 	is_current = self._current_project_key is None

		if ident is None:
			if self._current_project_key is not None:
				return self._loaded_projects[self._current_project_key]

			ident = self._infer_project_path(ident)

		if ident in self._loaded_projects:
			return self._loaded_projects[ident]

		if isinstance(ident, ProjectBase):
			proj = ident
			ident = proj.data.get('name', None)
		else:
			# create new
			path = ident
			if ident in self.data.get('projects', {}):
				path = self.data['projects'][ident]

			if path is not None and not os.path.exists(path):
				raise self.UnknownProjectError(path)

			proj = self.Project(path, profile=self)
			proj = proj.validate()

		if 'name' not in proj.data and ident is not None:
			proj.data['name'] = ident.stem if isinstance(ident, Path) else ident

		if self._loaded_projects.get(proj.name, proj) is not proj:
			# raise ValueError(f'Project with name {proj.name} already exists!')
			prt.warning(f'project name already loaded: {proj.name} (will now overwrite)')

		if ident is not None:
			if not isinstance(ident, Path) and ident != proj.name:
				if self.data.get('projects', {}).get(proj.name) is not None:
					raise ValueError('project name already exists: %s', ident)
				prt.warning(f'project name does not match profiles name for it: {ident} != {proj.name}')

			if self._loaded_projects.get(ident, proj) is not proj:
				# raise ValueError(f'Project with name {proj.name} already exists!')
				prt.warning(f'project name already loaded: {ident} (will now overwrite)')
			self._loaded_projects[ident] = proj

		self._loaded_projects[proj.name] = proj
		if is_current:
			self._current_project_key = proj.name
		return proj







import os
from setuptools import setup, find_packages

info = {'__file__':os.path.join(os.path.abspath(os.path.dirname(__file__)), 'omnifig', '_info.py')}
with open(info['__file__'], 'r') as f:
	exec(f.read(), info)
del info['__file__']

if 'readme' in info:
	with open(info['readme'], 'r') as f:
		lines = f.readlines()
	
	readme = []
	valid = 'md' in info['readme']
	for line in lines:
		if valid:
			if 'end-setup-marker-do-not-remove' in line:
				valid = False
			else:
				readme.append(line)
		elif 'setup-marker-do-not-remove' in line:
			valid = True
	
	README = '\n'.join(readme)
else:
	README = ''

setup(name=info.get('name', None),
      version=info.get('version', None),
      description=info.get('description', None),
      long_description=README,
      # long_description_content_type=info.get('desc_content_type', 'text/x-rst'),
      url=info.get('url', None),
      author=info.get('author', None),
      author_email=info.get('author_email', None),
      license=info.get('license', None),
      packages=find_packages(),
      entry_points=info.get('entry_points', {}),
      install_requires=info.get('install_requires', []),
      zip_safe=info.get('zip_safe', False),
      )


# ```text
# @misc{omnifig,
# author = {Leeb, Felix},
# title = {Omni-fig: Configuration and Project Management for Python},
# year = {2020},
# publisher = {GitHub},
# journal = {GitHub repository},
# howpublished = {https://github.com/felixludos/omni-fig}
#  }
# ```



from typing import Optional, Any, Sequence, ContextManager
from omnibelt import Primitive, unspecified_argument

from ..abstract import AbstractConfig



class AbstractSearch:
	'''Abstract class for search objects used by the config object to find the format data'''
	def __init__(self, origin: AbstractConfig, queries: Optional[Sequence[str]], default: Optional[Any], **kwargs):
		super().__init__(**kwargs)


	class SearchFailed(AbstractConfig.SearchFailed):
		'''Raised when a search fails to find a value'''
		def __init__(self, *queries: str):
			super().__init__(', '.join(queries))
			self.queries = queries


	def find_node(self, silent: Optional[bool] = None) -> AbstractConfig:
		'''Finds the node that contains the product'''
		raise NotImplementedError


	def find_product(self, silent: Optional[bool] = None) -> Any:
		'''Finds the node that contains the product, and then extracts the product'''
		raise NotImplementedError


	@staticmethod
	def sub_search(origin: 'AbstractSearch') -> 'ContextManager':
		'''Creates a context manager for new searches to be able to reference the original search `origin`'''
		raise NotImplementedError



class AbstractReporter:
	'''Abstract class for reporters used by the config object to report changes'''
	@staticmethod
	def log(*msg, end='\n', sep=' ', silent=None) -> str:
		'''Prints the given message to the console'''
		raise NotImplementedError


	def get_key(self, trace: 'AbstractSearch') -> str:
		'''Returns the key of the node resulting of the search'''
		raise NotImplementedError


	def report_node(self, node: 'AbstractConfig', *, silent: bool = None) -> Optional[str]:
		'''Reports information about a config node'''
		raise NotImplementedError


	def report_product(self, node: 'AbstractConfig', *, silent: bool = None) -> Optional[str]:
		'''Reports the product of a config node'''
		raise NotImplementedError


	def report_default(self, node: 'AbstractConfig', default: Any, *, silent: bool = None) -> Optional[str]:
		'''Reports a config node defaulted to the given value'''
		raise NotImplementedError


	def report_iterator(self, node: 'AbstractConfig', product: Optional[bool] = False, *,
	                    silent: Optional[bool] = None) -> Optional[str]:
		'''Reports the start of an iterator over the config node'''
		raise NotImplementedError


	def reuse_product(self, node: 'AbstractConfig', product: Any, *, silent: bool = None) -> Optional[str]:
		'''Reports that the product of the given node is being reused'''
		raise NotImplementedError


	def create_primitive(self, node: 'AbstractConfig', value: Primitive = unspecified_argument, *,
	                     silent: bool = None) -> Optional[str]:
		'''Reports that the product of the given node is a primitive'''
		raise NotImplementedError


	def create_container(self, node: 'AbstractConfig', *, silent: bool = None) -> Optional[str]:
		'''Reports that the product of the given node is a container (e.g. a dict or list)'''
		raise NotImplementedError


	def create_component(self, node: 'AbstractConfig', *, component_type: str = None,
	                     modifiers: Optional[Sequence[str]] = None, creator_type: str = None,
	                     silent: bool = None) -> Optional[str]:
		'''Reports that the product of the given node is a component'''
		raise NotImplementedError









from typing import Union, Iterator, Sequence, Any, Optional, ContextManager
from pathlib import Path
from omnibelt import JSONABLE, unspecified_argument

from .abstract import AbstractProject, AbstractConfig
from .organization import get_profile



# region Projects
def get_current_project() -> AbstractProject:
	'''Get the current project, assuming a profile is loaded, otherwise returns None'''
	return get_profile().get_current_project()



def get_project(ident: Union[str, Path]  = None) -> AbstractProject:
	'''Checks the profile to return (and possibly load) a project given the name or path ``ident``'''
	return get_profile().get_project(ident)



def switch_project(ident: Union[str, Path] = None) -> AbstractProject:
	'''Switches the current project to the one of thegiven the project name or path ``ident``'''
	return get_profile().switch_project(ident)



def iterate_projects() -> Iterator[AbstractProject]:
	'''Iterate over all loaded projects'''
	return get_profile().iterate_projects()



def project_context(ident: Union[str, Path] = None) -> ContextManager:
	'''Context manager for switching to a project and then switching back'''
	return get_profile().project_context(ident)
# endregion

# region Running
def entry(script_name: Optional[str] = unspecified_argument) -> None:
	'''
	Recommended entry point when running a script from the terminal.
	This is also the entry point for the ``fig`` command.

	This collects the command line arguments in ``sys.argv`` and overrides the
	given script with ``script_name`` if it is provided

	Args:
		script_name: script to be run (maybe set with arguments) (overrides other arguments if provided)

	Returns:
		:code:`None`

	'''
	get_profile().entry(script_name=script_name)



def main(argv: Sequence[str], script_name: Optional[str] = unspecified_argument) -> Any:
	'''
	Runs the desired script using the provided ``argv`` which are treated as command line arguments

	Before running the script, this function initializes ``omni-fig`` using :func:`initialize()`,
	and then cleans up after running using :func:`cleanup()`.

	Args:
		argv: raw arguments as if passed in through the terminal
		script_name: name of registered script to be run (maybe set with arguments) (overrides other arguments if provided)

	Returns:
		The output of script that is run

	'''
	return get_profile().main(argv, script_name=script_name)



def run_script(script_name: str, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
	'''
	Runs the specified script registered with ``script_name`` using the current project.

	Args:
		script_name: Must be registered in the current project
		config: The config object passed to the script
		*args: Manual arguments to be passed to the script
		**kwargs: Manual keyword arguments to be passed to the script

	Returns:
		The output of the script, raises MissingScriptError if the script is not found

	'''
	return get_profile().run_script(script_name, config, *args, **kwargs)



def run(config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
	'''
	Runs the specified script registered with ``script_name`` using the current project.

	Args:
		config: The config object passed to the script
		*args: Manual arguments to be passed to the script
		**kwargs: Manual keyword arguments to be passed to the script

	Returns:
		The output of the script, raises MissingScriptError if the script is not found

	'''
	return get_profile().run(config, args=args, kwargs=kwargs)



def quick_run(script_name: str, *parents: str, **parameters: JSONABLE) -> Any:
	'''
	Convenience function to run a simple script without a given config object,
	instead the config is entirely created using the provided ``parents`` and ``parameters``.

	Args:
		script_name: name of registered script that is to be run
		*parents: any names of registered configs to load
		**parameters: any additional arguments to be provided manually

	Returns:
		The script output

	'''
	return get_profile().quick_run(script_name, *parents, **parameters)



def initialize(*projects: str, **settings: Any) -> None:
	'''
	Initializes omni-fig by running the "princeps" file (if one exists),
	loading the profile, and any active projects. Additionally, loads the
	project in the current working directory (by default).

	Generally, this function should be run before running any scripts, as it should register all
	necessary scripts, components, and configs when loading a project. It is automatically called
	when running the :func:`main()` function (ie. running through the terminal). However, when
	starting scripts from other environments (such as in a jupyter notebook), this should be called
	manually after importing ``omnifig``.

	Args:
		projects: additional projects that should be initialized
		settings: extra global settings (unused by default)

	Returns:
		:code:`None`

	'''
	return get_profile().initialize(*projects, **settings)
# endregion



# region Create Config
def create_config(*configs: str, **parameters: JSONABLE) -> AbstractConfig:
	'''
	Process the provided data to create a config object (using the current project).

	Args:
		configs: usually a list of parent configs to be merged
		parameters: any manual parameters to include in the config object

	Returns:
		Config object resulting from loading/merging `configs` and including `data`.
	'''
	return get_profile().create_config(*configs, **parameters)



def parse_argv(argv: Sequence[str], *, script_name: Optional[str] = None) -> AbstractConfig:
	'''
	Parses the given arguments and returns a config object.

	Arguments are expected in the following order (all of which are optional):
		1. Meta rules to modify the config loading process and run mode.
		2. Name of the script to run.
		3. Names of registered config files that should be loaded and merged (in order of precedence).
		4. Manual config parameters (usually keys, prefixed by :code:`--` and corresponding values)

	Args:
		argv: List of arguments to parse (expected to be :code:`sys.argv[1:]`).
		script_name: Manually specified name of the script (defaults to what is specified in the resulting config).

	Returns:
		Config object containing the parsed arguments.

	'''
	return get_profile().parse_argv(argv, script_name=script_name)
# endregion



from typing import Optional, Any

from .. import __logger__ as prt
from ..abstract import AbstractConfig, AbstractProject

from .base import Behavior



class Debug(Behavior, name='debug', code='d', priority=100, description='Switch to debug mode'):
	'''
	When activated, this behavior updates the config object with the config file ``debug``.
	If for any reason the config file ``debug`` has already been loaded, this behavior will do nothing.

	Note that only a local ``debug`` config is merged, so the ``debug`` config file must be
	registered with the current project.
	'''
	def __init__(self, project: AbstractProject, **kwargs: Any):
		'''
		Sets up the debug behavior, including setting up a flag to make sure the debug config is only loaded once.

		Args:
			project: used to create the debug config
			**kwargs: passed to super
		'''
		super().__init__(project=project, **kwargs)
		self.project = project
		self._debug_done = False


	def pre_run(self, meta: AbstractConfig, config: AbstractConfig) -> Optional[AbstractConfig]:
		'''
		When activated, this behavior updates the config object with the config file ``debug``.

		If for any reason the config file ``debug`` has already been loaded, this behavior will do nothing.

		Args:
			meta: The meta config object (used to check if the behavior is activated with the ``debug`` key)
			config: Config object which will be passed to the script.

		Returns:
			the config object possibly updated with the ``debug`` config file

		Raises:
			:exc:`ConfigNotFoundError`: if the config file ``debug`` does not exist

		'''
		if meta.pull('debug', False, silent=True) and not self._debug_done:
			self._debug_done = True
			entry = self.project.find_local_artifact('config', 'debug', default=None)
			if entry is not None:
				debug = self.project.create_config(entry.path)
				cro = ('debug',) + debug.cro[1:] + tuple(c for c in config.cro if c not in debug.cro)
				config.update(debug)
				config._cro = cro
				config._bases = ('debug',)
				prt.info('Using debug mode')
				return config

			prt.error('No debug config found')



